with Ada.Text_IO, Strings, Rationals, Flts, Globals, Param, Reps;
use Ada.Text_IO, Strings, Rationals, Flts, Globals, Param, Reps;
with Matrices, Vectors, ScalVectors, Fouriers2, Fouriers2.Beam;

pragma Elaborate_All (Ada.Text_IO,Strings,Rationals,Flts,Globals,Param,Reps);
pragma Elaborate_All (Matrices,Vectors,ScalVectors,Fouriers2,Fouriers2.Beam);

procedure Approx_Fix(V: in Pars) is
  
  subtype Scalar is Rep;
  HPos:  constant Boolean  := HPos_Std(V);

  package SM is new Matrices (Scalar => Scalar);
  package SV is new Vectors (Component => Scalar);
  package SSV is new ScalVectors (Scalar => Scalar, SM => SM, SV => SV);
  package SFouriers2 is new Fouriers2 (Deg1 => V.Deg1, Deg2 => V.Deg2, Scalar => Scalar, SM => SM, SV => SV);
  package SFouriers2_Beam is new SFouriers2.Beam (HPos => HPos, PDeg1 => V.PDeg1, PDeg2 => V.PDeg2, SSV => SSV);
  use  SM, SFouriers2, SFouriers2_Beam, SFouriers2_Beam.Contr;
  
  PostFix: constant String := "fix_" & Fix_Name(V);
  PostMat: constant String := "mat_" & Fix_Name(V) & "." & Strng0(VDim,3);
  
  procedure Rough_Fix(F: in out Fourier2) is
    Fac: constant Flt := 1.33;
  begin
    SetZero(F);
    F.T.R := (V.Rho1,V.Rho2);
    F.T.P := (0,1);
    F.C(1,1) := Fac*Sqrt(Scal(abs(Sqr(V.Alpha)-Sqr(V.Gamma2)-V.Kappa)));
  end Rough_Fix;
  
  procedure Last_Fix(F: in out Fourier2) is
  begin
    Read(Tmp_Dir & "num" & "fix_last",F);
  exception
    when Name_Error => Rough_Fix(F);
  end Last_Fix;
  
  procedure Read_Fix(F: in out Fourier2) is
  begin
    Read(Input_Dir & "num" & PostFix,F);
  exception
    when Name_Error => Rough_Fix(F);
  end Read_Fix;
  
  procedure Write_Fix(F: in Fourier2) is
  begin
    Write(Tmp_Dir & "num" & "fix_last",F);
    Write(Output_Dir & "num" & PostFix,F);
  end Write_Fix;
  
  procedure Read_Mat(A: in out PMat) is
  begin
    Read(Tmp_Dir  & PostMat,A);
  exception
    when Name_Error => SetZero(A);
  end Read_Mat;
  
  procedure Write_Mat(A: in PMat) is
  begin
    Write(Tmp_Dir & PostMat,A);
  end Write_Mat;
  
  OK: Boolean;
  A: PMat;
  F: Fourier2;
  
begin
  Set_Precision(Num_Bits);
  FProd_Parallel := True;

  Read_Mat(A);
  Read_Fix(F);
  --  SetZero(A);
  --  Rough_Fix(F);
  
  Find_Fix(V,F,A,OK);
  if OK then
    Write_Mat(A);
    Write_Fix(F);
  end if;
  
end Approx_Fix;
