with Ada.Text_IO, Strings, Ints, Flts, Globals, Reps, Rationals, Flt_Vectors, Param;
use Ada.Text_IO, Strings, Ints, Flts, Globals, Reps, Rationals, Flt_Vectors, Param;
with Matrices, Vectors, ScalVectors, Fouriers2,Fouriers2.FlokM;

pragma Elaborate_All (Ada.Text_IO,Strings,Ints,Flts,Globals,Reps,Rationals,Flt_Vectors,Param);
pragma Elaborate_All (Matrices,Vectors,ScalVectors,Fouriers2,Fouriers2.FlokM);

with Std_Balls;
use Std_Balls;
pragma Elaborate_All (Std_Balls);

procedure Barrier(V: in Pars; FreqParity: in Parity) is

   subtype Scalar is Ball;

   package SM is new Matrices (Scalar => Scalar);
   package SV is new Vectors (Component => Scalar);
   package SSV is new ScalVectors (Scalar => Scalar, SM => SM, SV => SV);
   package SFouriers2 is new Fouriers2 (Deg1 => V.Deg1, Deg2 => V.Deg2, Scalar => Scalar, SM => SM, SV => SV);
   package SFouriers2_FlokM is new SFouriers2.FlokM (FreqParity=>FreqParity, DegHatM1=>V.DegHatM1, DegHat2=>V.DegHat2,
						     DegCheckM1=>V.DegCheckM1, DegCheckM2=>V.DegCheckM2, SSV => SSV);
   use SFouriers2,SFouriers2_FlokM,SM,SV,SSV;
   
   Postfix3:  constant String := Fix_Name(V);
   Postfix4:  constant String := Full_Name(V);
   Postfix3P: constant String := Postfix3 & ".P" & Strng(FreqParity);
   Postfix4P: constant String := Postfix4 & ".P" & Strng(FreqParity);
   Beta2:     constant Scalar := Scal(Sqr(V.Gamma1/V.Alpha));
   Sep:       constant FltVec := Read(Input_Dir & "sep-eta_" & Postfix4P,1);
   
   C: Scalar;
   H: Fourier2;
   D:  Flt_Vector(1..DegHatM2);
   
begin
   Set_Precision(256);
   MProd_Parallel := True;
   FProd_Parallel := True;
   Ops_Parallel := True;
   Show_Degs;
   
   Read(Input_Dir & "fixh_" & Postfix3,H);
   Read(Input_Dir & "c_" & Postfix3,C);
   Read(Input_Dir & "d_" & Postfix3P,D);
   Show1("C",C);
   Show1("|H| ",MaxNorm(H));
   if MaxNorm(H)>Inf(C) then raise Sorry with "||H||>C"; end if;
   CheckEta(Beta2,C,Sep,D,H);
   
end Barrier;
