with Messages, Strings, Ints, Rationals, Flts, Globals, Param, Reps, Reps.Balls, Approx_Fix;
use Strings, Ints, Rationals, Flts, Globals, Param, Reps, Reps.Balls;
with Matrices, Vectors, ScalVectors, Fouriers2, Fouriers2.Beam;

pragma Elaborate_All (Messages,Strings,Rationals,Flts,Globals,Param,Reps,Reps.Balls,Approx_Fix);
pragma Elaborate_All (Matrices,Vectors,ScalVectors,Fouriers2,Fouriers2.Beam);

procedure Check_Fix(V: in Pars) is

  subtype Scalar is Ball;

  Alpha: Rational renames V.Alpha;
  Gamma: Rational renames V.Gamma2;
  Kappa: Rational renames V.Kappa;
  HPos:  constant Boolean  := HPos_Std(V);

  package SM is new Matrices (Scalar => Scalar);
  package SV is new Vectors (Component => Scalar);
  package SSV is new ScalVectors (Scalar => Scalar, SM => SM, SV => SV);
  package SFouriers2 is new Fouriers2 (Deg1 => V.Deg1, Deg2 => V.Deg2, Scalar => Scalar, SM => SM, SV => SV);
  package SFouriers2_Beam is new SFouriers2.Beam (HPos => HPos, PDeg1 => V.PDeg1, PDeg2 => V.PDeg2, SSV => SSV);
  use  SM, SFouriers2, SFouriers2_Beam, SFouriers2_Beam.Contr;

  KMax:    constant Radius := Flt(3)/Flt(4);
  R0:     constant RadPair := (Zero,Zero);
  PostFix: constant String := "fix_" & Fix_Name(V);
  PostMat: constant String := "mat_" & Fix_Name(V) & "." & Strng0(VDim,3);
  
  procedure Make_Approx(F: in out Fourier2; A: in out PMat) is
    use Messages;
    Tr: constant Tracer := Starting("Check_Fix.Make_Approx");
  begin
    Reps.Proper_Rounding;
    Approx_Fix(V);
    NumRead(Input_Dir & "num" & PostFix,F);
    F.T.R := (V.Rho1,V.Rho2);
    NumRead(Tmp_Dir & PostMat,A);
    Reps.Balls.Proper_Rounding;
  end Make_Approx;
  
  A: PMat;
  F: Fourier2;
  Ok: Boolean := True;
  
begin
  Set_Precision(Num_Bits);
  Contr_Parallel := True;
  FProd_Parallel := True;

  SetEDegs(V.EDeg1,V.EDeg2);
  Show_Degs;
  Show1("alpha =",Alpha);
  Show1("gamma =",Gamma);
  Show1(" beta =",Gamma/Alpha);
  Show1("kappa =",Kappa);
  Show2("Rho  ",V.Rho1,V.Rho2);
  Show1("Precision ",Get_Precision(Reps.Pi));

  Make_Approx(F,A);
  CheckFix(V,F,A,KMax,OK);
  
  if OK then
    ProjOddFreq(F);
    Write(Output_Dir & PostFix,F);
    Show1("Check_Fix: done. norm0 =",Norm(R0,F));
  else
    Show0("Check_Fix: check failed");
  end if;
  
end Check_Fix;
