with Ada.Text_IO, Globals, Strings, Ints, Flts, Reps, Reps.Balls, Param, Flok_Intervals;
use Ada.Text_IO, Globals, Strings, Ints, Flts, Reps.Balls, Param, Flok_Intervals;
with Flt_Vectors, Matrices, Vectors, ScalVectors, ScalVectors.Phi;

pragma Elaborate_All (Ada.Text_IO,Globals,Strings,Ints,Flts,Reps.Balls,Param,Flok_Intervals);
pragma Elaborate_All (Matrices,Vectors,ScalVectors.Phi);

procedure Eigen_Circle(V: in Pars; FreqParity: in Parity; EigenBounds: in Boolean) is

   subtype Scalar is Ball;

   package Txt_IO renames Ada.Text_IO;
   package SM is new Matrices (Scalar => Scalar);
   package SV is new Vectors (Component => Scalar);
   package SSV is new ScalVectors (Scalar => Scalar, SM => SM, SV => SV);
   package SSV_Phi is new SSV.Phi;
   use SM,SV,SSV,SSV_Phi;

   Digs:      constant Positive := 768;
   DimX:      constant Positive := (V.DegHat2+FreqParity)/2;
   Dim2X:     constant Positive := 2*DimX;
   Eps:       constant Radius := 1.0E-28; --- for bisection in EigenCircle
   Postfix4P: constant String := Full_Name(V) & ".P" & Strng(FreqParity);
   ANameM:    constant String := "phi.D-." & Postfix4P;
   ANameP:    constant String := "phi.D+." & Postfix4P;
   
   procedure Eigenvectors_And_More(Name: in String) is
      A: Matrix(1..Dim2X,1..Dim2X);
      Wr,Wi: Vector(1..DimX);
      Vr,Vi: Matrix(1..Dim2X,1..DimX);
   begin
      Read(Input_Dir & Name,A);
      EigenCircle(A,Wr,Wi,Vr,Vi,Eps);
      PosIndexG(Wr,Wi,Vr,Vi);
      Plot_Eigen_Pos(Output_Dir & ANameM,Wr,Wi);
      Plot_Eigen_Neg(Output_Dir & ANameM,Wr,Wi);
      Plot_Eigen_Zero(Output_Dir & ANameM,Wr,Wi);
      if Name=ANameM then
	 declare
	    use Flt_Vectors;
	    Lo: constant F_Intervals := Low_Eta(Wr,Wi,Vr,Vi);
	    Hi: constant F_Intervals := Read(Input_Dir & "hi-eta_" & Postfix4P);
	    Sep: constant FltVec := Separating_Eta(Lo,Hi);
	 begin
	    Write(Output_Dir & "lo-eta_" & Postfix4P,Lo);
	    Write(Output_Dir & "sep-eta_" & Postfix4P,Sep);
	 end;
      end if;
      if Verbosity>4 then
	 Col_Normalize1(Vr,Vi);
	 EigenQuality(A,Wr,Wi,Vr,Vi);
      end if;
   end Eigenvectors_And_More;
   
begin
   
   Reps.Set_Precision(Digs);
   Eigenvectors_And_More(ANameM);
   if EigenBounds then
      Eigenvectors_And_More(ANameP);      
   end if;
   
end Eigen_Circle;
