with Strings, Globals, Quick_Sort;
use Strings, Globals;

package body Flok_Intervals is
  
  procedure Check_Intervals(V: in F_Intervals; Last: in Integer := Integer'Last) is
    Bad: Boolean;
    Den: Flt;
  begin
    for I in V'First .. IMin(Last,V'Last) loop
      declare
        VI: F_Interval renames V(I);
      begin
        Den := Flt(VI.Den);
        Bad := (VI.Num_Lower>VI.Num_Upper);
        if (VI.Num_Lower <= Zero) then
          if (VI.Num_Lower<Zero) or (VI.Index>0) then Bad := True; end if;
        end if;
        if (VI.Num_Upper >= Den) then
          if (VI.Num_Upper>Den) or (VI.Index<0) then Bad := True; end if;
        end if;
        if Bad then Show1(Strng(I,4),V(I)); raise F_Error; end if;
      end;
    end loop;
  end Check_Intervals;
  
  function Last_Nonempty(V: F_Intervals) return Integer is
  begin
    for L in reverse V'Range loop
      if V(L).Index /= 0 then return L; end if;
    end loop;
    return V'First-1;
  end Last_Nonempty;
  
  procedure Den1(V: in out F_Intervals) is
  begin
    for I in V'Range loop
      declare
        VI: F_Interval renames V(I);
      begin
        if VI.Den /= 1 then
          VI.Num_Lower := -(VI.Num_Lower/Flt(-VI.Den));
          VI.Num_Upper := VI.Num_Upper/Flt(VI.Den);
          VI.Den := 1;
        end if;
      end;
    end loop;
  end Den1;
  
  procedure Same_Den(V: in out F_Intervals) is
  begin
    if Last_Nonempty(V)<V'First then Den1(V); return; end if;  ------- ???????
    declare
      Done: Boolean := True;
      D: Positive := V(V'First).Den;
      Fac: Flt;
    begin
      for I in V'First+1 .. V'Last loop
        if V(I).Den /= D then
          D := LCM(D,V(I).Den);
          Done := False;
        end if;
      end loop;
      if Done then return; end if;
      for I in V'Range loop
        declare
          VI: F_Interval renames V(I);
        begin
          if VI.Den /= D then
            Fac := Flt(D/VI.Den);
            VI.Num_Lower := -(VI.Num_Lower*(-Fac));
            VI.Num_Upper := VI.Num_Upper*Fac;
            VI.Den := D;
          end if;
        end;
      end loop;
    end;
  end Same_Den;
  
  procedure Check_Overlaps(V: in F_Intervals; Last: in Integer := Integer'Last) is
--- use only after Sort  
  begin
    for I in V'First+1 .. IMin(Last,V'Last) loop
      if V(I).Num_Lower<V(I-1).Num_Upper then
        Show1(Strng(I-1,4),V(I-1));
        Show1(Strng(I,4),V(I));
        raise F_Error;
      end if;
    end loop;
  end Check_Overlaps;
  
  procedure Swap(A,B: in out F_Interval) is
    Tmp: constant F_Interval := A;
  begin
    A := B;
    B := Tmp;
  end Swap;
  
  function LT(A,B: F_Interval) return Boolean is
  begin
    if A.Index=0 then
      if B.Index /= 0 then return False; end if;
    elsif B.Index=0 then
      return True;
    end if;
    if A.Num_Lower<B.Num_Lower then return True; end if;
    if A.Num_Lower>B.Num_Lower then return False; end if;
    if A.Num_Upper<B.Num_Upper then return True; end if;
    if A.Num_Upper>B.Num_Upper then return False; end if;
    return False; -- should not really get here
  end LT;
  
  procedure Sort(V: in out F_Intervals; Check: in Boolean := True) is
    procedure QSort is new Quick_Sort (Item => F_Interval, Item_Vector => F_Intervals, "<" => LT);
  begin
    Same_Den(V);
    if Check then Check_Intervals(V); end if;
    QSort(V'First,V'Last,V);
  end Sort;
  
  procedure Try_Merge(A,B: in out F_Interval) is
--- assuming A.Den=B.Den
  begin
    if B.Index=0 then return; end if;
    if A.Index=0 then Swap(A,B); return; end if;
    if (A.Index>0)=(B.Index<0) then return; end if;
    if A.Num_Lower>B.Num_Lower then A.Num_Lower := B.Num_Lower; end if;
    if A.Num_Upper<B.Num_Upper then A.Num_Upper := B.Num_Upper; end if;
    A.Index := A.Index+B.Index;
    B.Index := 0;
    A.Finite := A.Finite and B.Finite;
  end Try_Merge;
  
  procedure Merge_PlusMinus(V: in out F_Intervals) is
    LNE,Last_LNE: Integer;
  begin
    Sort(V);
    LNE := Last_Nonempty(V);
    if LNE<V'First then return; end if;
    loop
      for I in reverse V'First+1 .. LNE loop
        Try_Merge(V(I-1),V(I));
      end loop;
      Sort(V,False);
      Last_LNE := LNE;
      LNE := Last_Nonempty(V);
      exit when LNE=Last_LNE;
    end loop;
    Check_Overlaps(V,LNE);
  end Merge_PlusMinus;
  
  function Trim_Empty(V: F_Intervals) return F_Intervals is
    L: Integer := 0;
    U: F_Intervals(1 .. V'Length);
  begin
    for I in V'Range loop
      if V(I).Index /= 0 then
        L := L+1;
        U(L) := V(I);
      end if;
    end loop;
    return U(1 .. L);
  end Trim_Empty;
  
  procedure Gap(Lower, Num_Upper, Den_Upper: in Radius; A: out F_Interval) is
  begin
    A.Num_Lower := Lower;
    if Den_Upper=One then
      A.Num_Upper := Num_Upper;
    else
      A.Num_Upper := -(Num_Upper/(-Den_Upper)); -- rounding down
    end if;
    A.Den := 1;
    A.Index := 0;
    A.Finite := True;
  end Gap;
  
  procedure Gaps(V: in out F_Intervals; L: out Integer; Z: out Parity) is
--- Z indicates whether there is a gap across zero (Z=1) or not (Z=0)
    FNE: constant Integer := V'First;
    LNE: Integer;
  begin
    Merge_PlusMinus(V);
    LNE := Last_Nonempty(V);
    if V(LNE).Index<0 and V(FNE).Index>0 then Z := 1; else Z := 0; end if;
    if Verbosity>3 then
      Show1("intervals",V(FNE .. LNE));
    end if;
    L := FNE-1;
    for I in V'First .. LNE-1 loop
      declare
        V0: F_Interval renames V(I);
        V1: F_Interval renames V(I+1);
      begin
        if V0.Index<0 and V1.Index>0 then
          L := L+1;
          Gap(V0.Num_Upper/Flt(V0.Den),V1.Num_Lower,Flt(V1.Den),V(L));
        end if;
      end;
    end loop;
    if Verbosity>3 then
      Show1("gaps",V(FNE .. L),True);
      if Z=1 then Show0("gap across zero"); end if;
    end if;
  end Gaps;
  
------------------------------------
  
  procedure Show1(N: in String; A: in F_Interval) is
  begin
    Show1(N & "Lower: ",A.Num_Lower,False);
    Show1(" /" & Strng(A.Den,3) & "  ~ ",A.Num_Lower/Flt(A.Den));
    Show1(N & "Upper: ",A.Num_Upper,False);
    Show1(" /" & Strng(A.Den,3) & "  ~ ",A.Num_Upper/Flt(A.Den));
    if A.Index=0 then
      Show0(N & "  Empty");
    elsif A.Finite then
      Show1(N & "  index",A.Index);
    elsif A.Index>0 then
      Show0(N & "  index +infinity");
    else
      Show0(N & "  index -infinity");
    end if;
  end Show1;
  
  procedure Show1(N: in String; V: in F_Intervals; ShowEmpty: Boolean := False) is
  begin
    for I in V'Range loop
      if (V(I).Index /= 0) or ShowEmpty then
        Show1(N & Strng(I,4),V(I));
      end if;
    end loop;
  end Show1;
  
  procedure Put(F: in File_Type; A: in F_Interval; Decimal: in Boolean := False) is
    Fin: constant Parity := Choose(A.Finite,1,0);
  begin
    Put(F,A.Num_Upper,Decimal);
    Put(F,A.Num_Lower,Decimal);
    Put(F,A.Den);
    Put(F,A.Index);
    Put(F,Fin);
  end Put;
  
  procedure Get(F: in File_Type; A: out F_Interval; Decimal: in Boolean := False) is
    pragma Warnings (Off); -- in out arguments in Get
    Fin: Parity;
  begin
    Get(F,A.Num_Upper,Decimal);
    Get(F,A.Num_Lower,Decimal);
    Get(F,A.Den);
    Get(F,A.Index);
    Get(F,Fin);
    A.Finite := (Fin=1);
  end Get;
  
  procedure Write(FileName: in String; V: in F_Intervals; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Writing " & FileName); end if;
    Create(F,Out_File,FileName);
    Put(F,V'Length);
    for I in V'Range loop
      Put(F,V(I),Decimal);
    end loop;
    Close(F);
  end Write;
  
  procedure Read(FileName: in String; V: in out F_Intervals; Decimal: in Boolean := False) is
    L: Integer;
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Reading " & FileName); end if;
    Open(F,In_File,FileName);
    Get(F,L);
    for I in V'First .. V'First+L-1 loop
      Get(F,V(I),Decimal);
    end loop;
    Close(F);
  end Read;
  
  function Read(FileName: String; Decimal: Boolean := False) return F_Intervals is
    L: Integer;
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Reading " & FileName); end if;
    Open(F,In_File,FileName);
    Get(F,L);
    declare
      V: F_Intervals(1 .. L);
    begin
      for I in 1 .. L loop
        Get(F,V(I),Decimal);
      end loop;
      Close(F);
      return V;
    end;
  end Read;
  
end Flok_Intervals;
