with Ada.Text_IO, Ints, Flts;
use Ada.Text_IO, Ints, Flts;

pragma Elaborate_All (Ada.Text_IO,Ints,Flts);

package Flok_Intervals is

  type F_Interval is
    record
      Num_Lower: Flt  := Zero; -- Num_Lower/Den is lower endpoint, <1
      Num_Upper: Flt  := One;  -- Num_Upper/Den is upper endpoint, >0
      Den: Positive   := 1;
      Index: Integer  := 0;    -- zero means empty interval
      Finite: Boolean := True; -- if False then |index| is infinite
    end record;

  type F_Intervals is array (Integer range <>) of F_Interval;

  procedure Check_Intervals(V: in F_Intervals; Last: in Integer := Integer'Last);
  function Last_Nonempty(V: F_Intervals) return Integer;
  procedure Den1(V: in out F_Intervals);
  procedure Same_Den(V: in out F_Intervals);
  procedure Sort(V: in out F_Intervals; Check: in Boolean := True);
  procedure Merge_PlusMinus(V: in out F_Intervals);
  function Trim_Empty(V: F_Intervals) return F_Intervals;

  procedure Gaps(V: in out F_Intervals; L: out Integer; Z: out Parity);
  procedure Show1(N: in String; A: in F_Interval);
  procedure Show1(N: in String; V: in F_Intervals; ShowEmpty: Boolean := False);
  procedure Put(F: in File_Type; A: in F_Interval; Decimal: in Boolean := False);
  procedure Get(F: in File_Type; A: out F_Interval; Decimal: in Boolean := False);
  procedure Write(FileName: in String; V: in F_Intervals; Decimal: in Boolean := False);
  procedure Read(FileName: in String; V: in out F_Intervals; Decimal: in Boolean := False);
  function Read(FileName: String; Decimal: Boolean := False) return F_Intervals;

  F_Error: exception;

private

  procedure Check_Overlaps(V: in F_Intervals; Last: in Integer := Integer'Last);
  procedure Swap(A,B: in out F_Interval);
  function LT(A,B: F_Interval) return Boolean;
  procedure Try_Merge(A,B: in out F_Interval);
  procedure Gap(Lower, Num_Upper, Den_Upper: in Radius; A: out F_Interval);

end Flok_Intervals;
