
package body Fouriers2.Beam is

  use SM,SV,SSV;
  
  procedure ProjOddFreq(F: in out Fourier2) is
  begin
    ProjOddFreq1(F);
    ProjOddFreq2(F);
  end ProjOddFreq;
  
  function DistInt(S: in Scalar) return Scalar is
    L: constant Integer := Integer(Flt'Floor(Approx(S)));
  begin
    return Min(S-Scal(L),Scal(L+1)-S);
  end DistInt;
  
  procedure InvLinear(V: in Pars; F: in out Fourier2) is
--- assumes Flt operations round up
    Alpha: Rational renames V.Alpha;
    Gamma: Rational renames V.Gamma2;
    Beta: constant Rational := Gamma/Alpha;
    Kappa: Rational renames V.Kappa;
    FC: FCoeff renames F.C;
    KGC: Rational;
  begin
    if OddOverOdd(Beta) then raise Not_Implemented; end if;
    if JustOdd then CleanOdd(F); end if;
    for K in Freq2 loop
      KGC := Sqr((K*K)*Gamma)+Kappa;
      for N in Freq1 loop
        if not IsZero(FC(N,K)) then
          Div(Scal(KGC-Sqr(N*Alpha)),FC(N,K));
        end if;
      end loop;
    end loop;
    if not STrunc then
      declare
        Q:  constant Positive := INum(Beta);
        P:  constant Positive := IDen(Beta);
        BC1: constant Scalar := Sqrt(Scal(Sqr(Beta)+Kappa/Sqr(Alpha))); -- beta*c_1
        Del1: constant Scalar := Flt(P)*DistInt(BC1);
        Del2: constant Rational := 1-(Kappa*Q)/(18*Sqr(Gamma));
        Del: constant Scalar := Min(Del1,Scal(Del2));
        DAP: constant Scalar := Del*Scal(Sqr(Alpha/P));    -- delta*(alpha/p)^2
        FE: FErr renames F.E;
      begin
	 if Inf(Del) <= Zero then raise Not_Implemented; end if;
	 for K in FreqE2 loop
	    for N in FreqE1 loop
	       if FE(N,K) /= Zero then
		  FE(N,K) := FE(N,K)/Inf((Scal(IMax(Q*K*K,P*N))-Del)*DAP);
	       end if;
	    end loop;
	 end loop;
      end;
    end if;
  end InvLinear;

  procedure MapN(V: in Pars; F: in Fourier2; G: in out Fourier2) is
  begin
    Prod(F,F,G);
    Mult(F,G);
    InvLinear(V,G);
    if HPos then Neg(G); end if;
  end MapN;

  procedure DMapN(V: in Pars; F: in Fourier2; H: in out Fourier2) is
  begin
    Mult(F,H);
    Mult(F,H);
    InvLinear(V,H);
    if HPos then
      Mult(-Three,H);
    else
      Mult(Three,H);
    end if;
  end DMapN;

  ------------ numerical

  procedure Find_Fix(V: Pars; F: in out Fourier2; A: in out PMat; OK: out Boolean) is
    Eps:  constant Radius := 1.0e-15; --- first reasonable precision
    Eps0: constant Radius := 1.0e-12;
    Err,Nf: Radius;
    G: Fourier2;
  begin
    if Verbosity>1 then
      Show1("alpha = ",V.Alpha);
      Show1("gamma = ",V.Gamma2);
      Show1("kappa = ",V.Kappa);
      Nf := MaxNorm(F);
      MapN(V,F,G);
      Sub(F,G);
      Err := MaxNorm(G);
      Show1("old Err =",Err,False);
      Show1("   norm =",Nf);
    end if;
    if IsZero(A) then
      ApproxFix(V,F,A,Eps);
    else
      IterContr(V,F,A,Eps);
    end if;
    MapN(V,F,G);
    Sub(F,G);
    Err := MaxNorm(G);
    if Verbosity>0 then
      Nf := MaxNorm(F);
      Show1("new Err =",Err,False);
      Show1("   norm =",Nf);
    end if;
    OK := (Err<Eps0);
  end Find_Fix;

end Fouriers2.Beam;
