with ScalVectors, ScalVectors.RNum;

pragma Elaborate_All (ScalVectors.RNum);
   
generic
  
  PDeg1,PDeg2: in Positive;    --- pack degres, choose odd if JustOdd is true
  JustOdd: in Boolean := True; --- pack only (odd,odd) frequencies
  
  type Param is private;
  
  with package SSV is new ScalVectors (Scalar => Scalar, SM => SM, SV => SV);
  
  with procedure Map(P: in Param; F: in Fourier2; G: in out Fourier2) is <>;
  with procedure DMap(P: in Param; F: in Fourier2; H: in out Fourier2) is <>;
  
package Fouriers2.Contr is
  
  use SM,SV,SSV;
  
  VDim: constant Integer := NumCoeffs(PDeg1,PDeg2,JustOdd);
  
  subtype VRange is Integer range 1 .. VDim;
  subtype PVec is SV.Vector(VRange);
  subtype PMat is SM.Matrix(VRange,VRange);
  
  procedure Show_Degs;
  procedure Contract(P: in Param; F: in Fourier2; A: in PMat; H: in out Fourier2; Err: out Radius);
  procedure SetEDegs(EDeg1,EDeg2: in Positive);
  function DContractNorm(P: Param; F: Fourier2; A: PMat; EDeg1,EDeg2: Positive) return Radius;
  function DContractNorm(P: Param; F: Fourier2; A: PMat) return Radius;
  procedure CheckFix(P: in Param; F: in out Fourier2; A: in PMat; KMax: in Radius; OK: out Boolean);
  procedure CheckFix(P: in Param; F: in out Fourier2; A: in PMat; KMax: in Radius);
  procedure CheckFix(P: in Param; F: in out Fourier2; KMax: in Radius);
  
--- numerical
  package SSV_RNum is new SSV.RNum;
  
  procedure DerMatrix(P: in Param; F: in Fourier2; A: in out PMat);
  procedure ContrMatrix(P: in Param; F: in Fourier2; A: in out PMat);
  procedure IterContr(P: in Param; F: in out Fourier2; A: in out PMat; Eps: in Radius);
  procedure ApproxFix(P: in Param; F: in out Fourier2; A: in out PMat; Eps: in Radius);
  procedure ApproxFix(P: in Param; F: in out Fourier2; Eps: in Radius);
  procedure Eigen(P: in Param; F: in Fourier2; Wr: in out PVec; Large: out Integer);
  procedure Eigen(P: in Param; F: in Fourier2; Large: out Natural);
  procedure NumRead(Name: in String; A: in out PMat; Decimal: in Boolean := False);
  
  Contr_Parallel: Boolean := False; --- used by DContractNorm and DerMatrix
  
private
  
  function NoLowErrs(F: Fourier2) return Boolean;
  procedure DistErr(F: in out Fourier2);
  procedure Pack(C: in FCoeff; V: in out PVec);
  procedure Pack(F: in Fourier2; V: in out PVec);
  procedure UnPack(V: in PVec; C: in out FCoeff);
  procedure UnPack(V: in PVec; T: in FType; F: in out Fourier2);
  procedure Prod(A: in PMat; F: in Fourier2; G: in out Fourier2);
  function OpNorm(T: FType; M: PMat) return Radius;
  function Check_EDeg1(D: Positive) return Positive;
  function Check_EDeg2(D: Positive) return Positive;
  procedure Show_Low(ID,I,J: in Integer; R: in Radius);
  procedure Show_High(ID,I,J: in Integer; R: in Radius);
  
  protected type Protected_Ops is
    procedure Initialize(EDeg1,EDeg2: in Positive);
    procedure New_Low_Mode(I,J: out Integer; Done: out Boolean);
    procedure New_High_Mode(I,J: out Integer; Done: out Boolean);
    procedure Check_Max(ID,I,J: in Integer; R: in Radius);
    procedure Get_Max(R: out Radius);
  private
    ED1,ED2,I0,J0,ECount: Integer;
    R0: Radius := Zero;
  end Protected_Ops;
  
  Lowest:    constant Integer  := Choose(JustOdd,1,0); --- lowest index
  Increment: constant Integer  := Choose(JustOdd,2,1);
  
  Default_EDeg1: Positive := PDeg1+Increment;
  Default_EDeg2: Positive := PDeg2+Increment;
  
  Prot: Protected_Ops;
  
end Fouriers2.Contr;
