with Strings;
use Strings;

package body Fouriers2.FlokD is
  
  procedure Show_Degs is
  begin
    Show_Param_F2("Fouriers2 ");
    Show1("Fouriers2.FlokD FreqParity =",FreqParity);
    Show2("Fouriers2.FlokD DegHat1 DegHat2 =",DegHat1,DegHat2);
  end Show_Degs;
  
  function Weight(J: Integer; K: Positive) return Radius is
  begin
    if abs(J)>DegHat2 then return Half; end if;
    if J=0 then return Zero; end if;
    if K>DegHat2 then return Half; end if;
    return Zero;
  end Weight;
  
  function MakeD(C: FCoeff; K: Positive) return Radius is
    D,R: Radius := Zero;
  begin
    for J in Freq2 loop
      R := MaxNorm(C(0,J));
      if R /= Zero then
        D := D+R*(Weight(J+K,K)+Weight(J-K,K));
      end if;
    end loop;
    return D;
  end MakeD;
  
  procedure SplitErr0(F: in out Fourier2; E: out Radius) is
    --- sum all errors and all N>DegHat1 terms into E
    --- resulting F.C is sharp, and F.E is zero
    FC: FCoeff renames F.C;
    FE: FErr renames F.E;
    S: Scalar;
  begin
    if DegHat1>Deg1 then raise Not_Implemented; end if;
    E := Zero;
    if STrunc then return; end if;
    for N in reverse FreqE1 loop
      for K in reverse FreqE2 loop
        E := E+FE(N,K);
        FE(N,K) := Zero;
      end loop;
    end loop;
    for N in reverse DegHat1+1 .. Deg1 loop
      for K in reverse Freq2 loop
        E := E+MaxNorm(FC(N,K));
        SetZero(FC(N,K));
      end loop;
    end loop;
    for N in reverse 0 .. DegHat1 loop
      for K in reverse Freq2 loop
        ModCenter(FC(N,K),S);
        E := E+MaxNorm(S);
        Center(FC(N,K));
      end loop;
    end loop;
  end SplitErr0;
  
  procedure TSup(GC: in FCoeff; C: in out FCoeff) is
    use SM;
    CK: Radius;
  begin
    SetZero(C);
    for K in Freq2 loop
      CK := Zero;
      for N in reverse Freq1 loop
        CK := CK+MaxNorm(GC(N,K));
      end loop;
      BallAt0(CK,C(0,K));
    end loop;
  end TSup;
  
  procedure MakeD(G: in out Fourier2; D: out Flt_Vector) is
    L: constant Integer := D'Last;
    E: Flt;
    
    procedure SaveD(K: in Positive; R: in Radius) is
    begin
      D(K) := R+E;
      if Verbosity>1 then Show1("D(" & Strng(K,2) & ") =",D(K)); end if;
    end SaveD;
    
    K: Natural;
    R: Radius;
    C: FCoeff;
  begin
    ProjEvenFreq1(G);
    ProjEvenFreq2(G);
    SplitErr0(G,E);
    if Verbosity>1 then Show1("E =",E); end if;
    TSup(G.C,C);
    R := MakeD(C,Deg2+1);
    for I in reverse L .. Deg2 loop
      K := KOffset+2*I;
      R := RMax(R,MakeD(C,K));
    end loop;
    SaveD(L,R);
    for I in reverse 1 .. L-1 loop
      K := KOffset+2*I;
      R := MakeD(C,K);
      SaveD(I,R);
    end loop;
  end MakeD;
  
-----------------------------
  
  procedure Make_Neg(Lower: in Rational; Upper: in Radius; A: out F_Interval) is
--- negative-index eta_k (mod 1) accumulating at Lower
--- called with Lower and Upper in [0,1]
  begin
    A.Den := IDen(Lower);
    A.Num_Lower := Flt(INum(Lower));
    A.Num_Upper := Flt(A.Den)*Upper; -- rounded up
    A.Index := -1;
    A.Finite := False;
  end Make_Neg;
  
  procedure Make_Pos(Lower: in Flt; Upper: in Rational; A: out F_Interval) is
--- positive-index eta_k (mod 1) accumulating at Upper
--- called with Lower and Upper in [-1,0], so add 1
  begin
    A.Den := IDen(Upper);
    A.Num_Lower := -(Flt(A.Den)*(NegOne-Lower)); -- rounded down
    A.Num_Upper := Flt(INum(1+Upper));
    A.Index := 1;
    A.Finite := False;
  end Make_Pos;
  
  function High_Eta(Beta: in Rational; C: in Scalar; DBound: in Radius) return F_Intervals is
    Ck: constant Scalar := C+Scal(DBound); -- C should include G.C(0)
    Yk: Scalar;
    
    procedure NegIndex_Eta(K: in Positive; Lower: out Rational; Upper: out Radius) is
    --- [Lower,Upper] contains eta_k
    begin
      Yk := Sqrt(Scal(Sqr(K*Beta))+Ck);
      Lower := (K*Beta) mod 1;               -- accumulation point
      Upper := Sup(Yk-Scal(K*Beta-Lower));
    end NegIndex_Eta;
    
    K: Natural := KFirst;
    L: Natural := 0;
    Lower,E: Rational;
    Upper: Radius;
    V: F_Intervals(1 .. 2*IDen(Beta));
  begin
    loop
      if K>DegHat2 then
        NegIndex_Eta(K,Lower,Upper);
        if L=0 then E := Lower; elsif Lower=E then exit; end if;
        L := L+1; Make_Neg( Lower, Upper,V(L));
        L := L+1; Make_Pos(-Upper,-Lower,V(L));
      end if;
      K := K+2;
    end loop;
    Check_Intervals(V,L);
    if Verbosity>4 then Show1("high_eta",V); end if;
    return V(1 .. L);
  end High_Eta;
      
end Fouriers2.FlokD;
