with Strings;
use Strings;

package body Fouriers2.FlokM is
  
  procedure Show_Degs is
  begin
    Show_Param_F2("Fouriers2 ");
    Show1("Fouriers2.FlokM FreqParity =",FreqParity);
    Show2("Fouriers2.FlokM DegHatM1 DegHatM2 =",DegHatM1,DegHatM2);
    Show2("Fouriers2.FlokM DegCheckM1 DegCheckM2 =",DegCheckM1,DegCheckM2);
    Show2("Fouriers2.FlokM HDim CDim =",HDim,CDim);
  end Show_Degs;
  
  procedure Show1(S: in String; F: in TwoFourier2) is
  begin
    Show1(S,F(0));
    Show1(S,F(1));
  end Show1;
  
  procedure LiftErrs2(F: in out Fourier2; D1,D2: in Natural) is
--- the new F.E(*,0..D2) are independent of the second variable
--- and F.E(0..D1,0..D2)=0
    W1: constant Flt := NotExp(-F.T.R(1));
    W2: constant Flt := NotExp(-F.T.R(2));
    FC: FCoeff renames F.C;
    FE: FErr renames F.E;
    Fac1,Fac2: Flt;
    B: Scalar;
  begin
    if STrunc then return; end if;
    CleanOdd(F);
    Separate2(F.E,DegE1,D2); -- FE(*,0..D2) are now independent of the second variable
    Separate1(F.E,D1,D2);    -- FE(0..D1,0..D2) are now constant and ...
    Fac1 := One;
    for N in 0 .. D1 loop    -- we include them in FC(0..D1,0..D2)
      Fac2 := Fac1;
      for K in 0 .. D2 loop
        BallAt0(FE(N,K)*Fac2,B);
        Add(B,FC(N,K));
        FE(N,K) := Zero;
        Fac2 := W2*Fac2;
      end loop;
      Fac1 := W1*Fac1;
    end loop;
  end LiftErrs2;
  
  procedure Concentrate2(F: in out Fourier2; D1,D2: in Natural) is
--- for K in 0..D2 the new F has
--- F.C(N,K)=0 for N>D1 and F.E(N,K)=0 for N=/=D1+1
--- with F.E(D1+1,K) independent of the first variable
    D11: constant Positive := D1+1;
    FC: FCoeff renames F.C;
  begin
    if STrunc then
      for K in 0 .. D2 loop
        for N in D11 .. Deg1 loop SetZero(FC(N,K)); end loop;
      end loop;
    else
      LiftErrs2(F,D1,D2);
      declare
        D12: constant Positive := D1+2;
        W1: constant Flt := NotExp(F.T.R(1));
        W2: constant Flt := NotExp(F.T.R(2));
        FE: FErr renames F.E;
        Fac2: Radius := Power(W1,D11);
        R: Radius;
      begin
        for K in 0 .. D2 loop
          R := Zero;
          for N in reverse D11 .. Deg1 loop
            R := W1*R+MaxNorm(FC(N,K));
            SetZero(FC(N,K));
          end loop;
          declare
            E: Radius renames FE(D11,K);
          begin
            for N in reverse D12 .. DegE1 loop
              E := E+FE(N,K);
              FE(N,K) := Zero;
            end loop;
            E := E+R*Fac2;
          end;
          Fac2 := W2*Fac2;
        end loop;
      end;
    end if;
  end Concentrate2;
  
  procedure Pack(F: in TwoFourier2; V: in out HVector) is
    Tmp: LF_Pointer2;
  begin
    Pool2.Allocate(Tmp);
    declare
      G: TwoFourier2 renames Tmp.Data;
    begin
      for P in Parity loop
        Copy(F(P),G(P));
        ProjFreq2(FreqParity,G(P));
        Concentrate2(G(P),DegHatM1,DegHatM2);
      end loop;
      for J in HRange loop
        declare
          IndexJ: Monom renames Index(J);
          P: Parity renames IndexJ.P;
          N: Integer renames IndexJ.N;
          K: Integer renames IndexJ.K;
        begin
          if IndexJ.E then
            BallAt0(G(P).E(N,K),V(J));
          else
            Copy(G(P).C(N,K),V(J));
          end if;
        end;
      end loop;
    end;
    Pool2.Recycle(Tmp);
  end Pack;

  procedure UnPack(V: in HVector; Template: in TwoFourier2; G: in out TwoFourier2) is
  begin
    SetZero(Template(0).T,G(0));
    SetZero(Template(1).T,G(1));
    for J in HRange loop
      declare
        IndexJ: Monom renames Index(J);
        N: Integer renames IndexJ.N;
        K: Integer renames IndexJ.K;
        P: Parity renames IndexJ.P;
      begin
        if IndexJ.E then
          G(P).E(N,K) := MaxNorm(V(J));
        else
          Copy(V(J),G(P).C(N,K));
        end if;
      end;
    end loop;
  end UnPack;

--------------------------------------  
  
  procedure Theta(F: in out Fourier2) is
    -- assumes rounding up for Flt ops
    -- truncates K>DegHatM2
    FC: FCoeff renames F.C;
    FE: FErr renames F.E;
  begin
    CleanOdd2(F);
    for K in Freq2 loop
      if K>DegHatM2 then
        for N in Freq1 loop SetZero(FC(N,K)); end loop;
      elsif (K mod 2)=FreqParity then
        for N in Freq1 loop Div(Scal(IMax(N+K,1)),FC(N,K)); end loop;
      end if;
    end loop;
    if STrunc then return; end if;
    for K in FreqE2 loop
      if K>DegHatM2 then
        for N in FreqE1 loop FE(N,K) := Zero; end loop;
      elsif (K mod 2)=FreqParity then
        for N in FreqE1 loop FE(N,K) := FE(N,K)/Flt(IMax(N+K,1)); end loop;
      else
        for N in FreqE1 loop
          FE(N,K+1) := FE(N,K+1)+FE(N,K);
          FE(N,K):= Zero;
        end loop;
      end if;
    end loop;
  end Theta;
  
  procedure THT(H: in Fourier2; F: in TwoFourier2; G: in out TwoFourier2) is
  begin
    for P in Parity loop
      Copy(F(P),G(P));
      Theta(G(P));
      Mult(H,G(P));
      Theta(G(P));
    end loop;
  end THT;
  
  function Norm(F: TwoFourier2) return Scalar is
  begin
    return NormOrRadius(F(0))+NormOrRadius(F(1));
  end Norm;

  function THT_Norm(H: Fourier2; Template: TwoFourier2) return Radius is
    procedure Op(F: in TwoFourier2; G: in out TwoFourier2) is begin THT_Op(H,F,G); end Op;
    function Op_NormBound is new OON.OpNormBound (Op => Op);
  begin
    return Op_NormBound(Template,Ops_Parallel);
  end THT_Norm;

  procedure THT_Matrix(H: in Fourier2; Template: in TwoFourier2; THT: in out HMatrix) is
    procedure Op(F: in TwoFourier2; G: in out TwoFourier2) is begin THT_Op(H,F,G); end Op;
    procedure PO is new OON.Packed_Op (Op => Op);
  begin
    PO(Template,THT,Ops_Parallel);
  end THT_Matrix;
  
  function HatMinusCheckMaxNorm2(THT: HMatrix) return Radius is
    Pow4: constant Positive := 3;
    R: Radius;
    S: Scalar;
    HMC,Tmp: HMatrix;
  begin
    Copy(THT,HMC);
    for I in CRange loop
      for J in CRange loop
        ModCenter(HMC(I,J));
      end loop;
    end loop;
    MaxEigen(HMC,Tmp,S,Pow4);
    R := Sup(S);
    if Verbosity>1 then
      Show1("|hatH-checkH| <",R);
    end if;
    return R;
  end HatMinusCheckMaxNorm2;
  
------------------------------
  
  function TMT_High(Beta2,C,Eta: Scalar; D: Flt_Vector) return Scalar is --- check this
--- just a lower bound
    N: constant Integer := DegCheckM1+1;
    K: Integer := KFirst;
    S: Scalar := SOne;
    B: Scalar;
  begin
    if Sup(Eta)>=One then raise Not_Implemented; end if;
    while K <= DegHatM2 loop
      B := (Sqr(Scal(N)+Eta)-Beta2*Scal(K*K)-C-Scal(D(K)))/Scal((N+K)*(N+K));
      S := Min(S,B);
      K := K+2;
    end loop;
    if Inf(S)<=Zero then raise Sorry with "increase DegCheckM1"; end if;
    return S;
  end TMT_High;
  
  procedure TMT_Low(Beta2,C,Eta,S: in Scalar; D: in Flt_Vector; THT: in HMatrix; TMT: in out CMatrix) is
--- used to be MakeM
    EC: constant Scalar := Sqr(Eta)-C;
    NK,I: Integer;
    Fac,B: Scalar;
  begin
    for J in CRange loop
      for I in CRange loop
        Center(THT(I,J),TMT(I,J));
      end loop;
      if not Index(J).E then
        declare
          P: Parity renames Index(J).P;
          N: Integer renames Index(J).N;
          K: Integer renames Index(J).K;
        begin
          NK := IMax(N+K,1);
          Fac := Inv(Scal(NK*NK));
          B := Fac*(Scal(N*N)-Beta2*Scal(K*K)+EC-D(K)*S);
          Add(B,TMT(J,J));
          if N>0 then
            B := Fac*Eta*Scal(2*N);
            I := Find(Index,1-P,N,K);
            Add(B,TMT(I,J));
          end if;
        end;
      end if;
    end loop;
  end TMT_Low;
  
  function CNorm(D: Flt_Vector) return Radius is
    K: Natural := KFirst;
    R: Radius := Zero;
  begin
    for I in Positive loop
      R := RMax(R,abs(D(I))/Flt(IMax(K*K,1)));
      K := K+2;
      exit when K>DegCheckM2;
    end loop;
    return R;
  end CNorm;
  
  procedure SpecGap(A,Tmp: in out CMatrix; Need: in Radius; OK: out Boolean) is
--- modifies A and Tmp
     Pow2: constant Positive := 8;
     Fac: Flt := NegOne;
     Gap: Radius;
  begin
     Prod(A,A,Tmp);
     OK := True;
     AddConst(Scal(-Need*Need),Tmp);
     if IsPositive(Tmp)=True then
	Show0("gap test1 OK.");
     else
	Show0("gap test2 ...");
	Invert(A);
	for K in 1 .. Pow2 loop
	   Prod(A,A,Tmp);
	   Fac := Half*Fac;
	   Gap := Inf(Exp(Fac*Log(Norm1(Tmp))));
	   if Gap>Need then
	      Show2("OK. gap, need:",Gap,Need);
	      return;
	   end if;
	   Swap(A,Tmp);
	end loop;
	Show2("gap less than needed:",Gap,Need);
	OK := False;
     end if;
  end SpecGap;

  procedure CheckEta(Beta2,C,Eta: in Scalar; D: in Flt_Vector; THT: in HMatrix; Need: in Radius) is
    DS: constant Flt := Need*(-Two)/CNorm(D);
    OK: Boolean;
    Gap: Radius;
    TMT,Tmp: CMatrix;

    procedure CheckGap(S: in Flt) is
    begin
      TMT_Low(Beta2,C,Eta,Scal(S),D,THT,TMT);
      SpecGap(TMT,Tmp,Need,OK);
      if not OK then raise Sorry; end if;
    end CheckGap;

    S: Flt := DS;
  begin
     if Verbosity>3 then Show1("Checking eta=",Eta); end if;
     Gap := Inf(TMT_High(Beta2,C,Eta,D));
     if Gap<Need then
	Show2("high gap less than needed: ",Gap,Need); raise Sorry;
     end if;
     CheckGap(Zero);
     while S>NegOne loop
	CheckGap( S);
	CheckGap(-S);
	S := S+DS;
     end loop;
     CheckGap(-One);
     CheckGap( One);
  end CheckEta;
  
  procedure CheckEta(Beta2,C: in Scalar; Eta,D: in Flt_Vector; H: in Fourier2) is
    Eps: constant Radius := 1.0E-99;
    Need: Radius;
    Tmp: TwoFourier2;
    THT: HMatrix;
  begin
    for P in Parity loop
      Tmp(P).T.P := (P,FreqParity);
      Tmp(P).T.R := H.T.R;
    end loop;
    THT_Matrix(H,Tmp,THT);
    Need := HatMinusCheckMaxNorm2(THT)+Eps;
    for I in Eta'Range loop
       CheckEta(Beta2,C,Scal(Eta(I)),D,THT,Need);
    end loop;
    Show0("CheckEta: done. checked the " & Strng(Eta'Length) & " separating values");
  end CheckEta;
  
  procedure Free_Cache_TF2 is
  begin
    Pool2.Free_All;
    Free_Cache_F2;
  end Free_Cache_TF2;
  
begin
  
  if DegHatM1>Deg1 then raise Sorry with "DegHatM1 too large"; end if;
  if DegHatM2>Deg2 then raise Sorry with "DegHatM2 too large"; end if;
  if DegCheckM1>DegHatM1 then raise Sorry with "DegCheckM1 too large"; end if;
  if DegCheckM2>DegHatM2 then raise Sorry with "DegCheckM2 too large"; end if;
  
end Fouriers2.FlokM;
