with Make_Index, ScalVectors, OpNorms;

generic
  
  FreqParity: in Parity;              -- even or odd x frequencies
  DegHatM1,DegHat2: in Positive;      -- for \hat subspace
  DegCheckM1,DegCheckM2: in Positive; -- for \check subspace
  
  with package SSV is new ScalVectors (Scalar => Scalar, SM => SM, SV => SV);

package Fouriers2.FlokM is
  
  DegHatM2: constant Positive := DegHat2;
  
  package MI is new Make_Index(Numeric => FTrunc, FreqParity => FreqParity,
                               DegHatM1 => DegHatM1, DegHatM2 => DegHatM2, DegCheckM1 => DegCheckM1);
  use SM,SV,SSV,MI;
  
  HDim: constant Positive  := Get_Count(False);
  subtype HRange is Integer range 1 .. HDim;    -- for \hat matrices
  subtype HVector is SV.Vector(HRange);
  subtype HMatrix is SM.Matrix(HRange,HRange);
  
  CDim: constant Positive  := Get_Count(True);
  subtype CRange is Integer range 1 .. CDim;    -- for \check matrices
  subtype CVector is SV.Vector(CRange);
  subtype CMatrix is SM.Matrix(CRange,CRange);
  
  type TwoFourier2 is array (Parity) of Fourier2;

  procedure Show_Degs;
  procedure Show1(S: in String; F: in TwoFourier2);
  procedure Pack(F: in TwoFourier2; V: in out HVector);
  procedure UnPack(V: in HVector; Template: in TwoFourier2; G: in out TwoFourier2);
  function MatNorm(Template: TwoFourier2; A: HMatrix) return Radius;

--- \hat stuff
  procedure Theta(F: in out Fourier2);
  procedure THT(H: in Fourier2; F: in TwoFourier2; G: in out TwoFourier2);
  function THT_Norm(H: Fourier2; Template: TwoFourier2) return Radius;
  procedure THT_Matrix(H: in Fourier2; Template: in TwoFourier2; THT: in out HMatrix);
  function HatMinusCheckMaxNorm2(THT: HMatrix) return Radius;
  
--- \check stuff
  function TMT_High(Beta2,C,Eta: Scalar; D: Flt_Vector) return Scalar;
  procedure TMT_Low(Beta2,C,Eta,S: in Scalar; D: in Flt_Vector; THT: in HMatrix; TMT: in out CMatrix); --was MakeM
  procedure CheckEta(Beta2,C,Eta: in Scalar; D: in Flt_Vector; THT: in HMatrix; Need: in Radius);
  procedure CheckEta(Beta2,C: in Scalar; Eta,D: in Flt_Vector; H: in Fourier2);
  
  procedure Free_Cache_TF2;
  
  Ops_Parallel: Boolean := False;
  
private
  
  KFirst: constant Positive := 2-FreqParity; -- first x-frequency
  Index: constant ParVector := Get_Index;
  
  procedure LiftErrs2(F: in out Fourier2; D1,D2: in Natural);
  procedure Concentrate2(F: in out Fourier2; D1,D2: in Natural);
  function Norm(F: TwoFourier2) return Scalar;
  procedure THT_op(H: in Fourier2; F: in TwoFourier2; G: in out TwoFourier2) renames THT;
  package OON is new OpNorms (Dim => HDim, Scalar => Scalar, Fun => TwoFourier2, SVector => SV.Vector, SMatrix => SM.Matrix);
  function MatNorm(Template: TwoFourier2; A: HMatrix) return Radius renames OON.NormBound;
  function CNorm(D: Flt_Vector) return Radius;
  procedure SpecGap(A,Tmp: in out CMatrix; Need: in Radius; OK: out Boolean);
  
  package Alloc2 is new Recycling (Item => TwoFourier2);
  subtype LF_Pointer2 is Alloc2.LD_Pointer;
  O2: Alloc2.Controlled_Protected_Data;
  Pool2: Alloc2.Protected_Data renames O2.P;
  
end Fouriers2.FlokM;
