with Strings, Protected_Counters;
use Strings, Protected_Counters;

pragma Elaborate_All (Strings,Protected_Counters);

package body Fouriers2 is

  use Flt_Matrices,Flt_Vectors,SM,SV,SV_More,SM_More;

  function NotExp(X: Radius) return Scalar is
--- 1+X
    S: Scalar;
  begin
    Assign(X,S);
    Add(SOne,S);
    return S;
  end NotExp;

  function NotExp(X: Flt) return Flt is
--- upper bound on 1+X if X>=0
--- upper bound on 1/(1-X) if X<0
  begin
    if X >= Zero then return One+X; end if;
    return NegOne/(X-One);
  end NotExp;

--- basic

  function Info(Dummy: Fourier2) return Scalar_Info is
    pragma Unreferenced(Dummy);
  begin
    return FInfo;
  end Info;

  function IsPoly(P: Fourier2) return Boolean is
  begin
    return STrunc or else IsZero(P.E);
  end IsPoly;

  function IsSharp(P: Fourier2) return Boolean is
  begin
    return STrunc or else (IsSharp(P.C) and then IsZero(P.E));
  end IsSharp;

  function IsZero(P: Fourier2) return Boolean is
  begin
    return IsZero(P.C) and then (STrunc or else IsZero(P.E));
  end IsZero;

  function "="(P1,P2: Fourier2) return Boolean is
    T1: FType renames P1.T;
    T2: FType renames P2.T;
  begin
    if T1.P /= T2.P then return False; end if;
    if (P1.C /= P2.C) then return False; end if;
    if STrunc then return True; end if;
    if (P1.E /= P2.E) then return False; end if;
    if IsZero(P1.E) then return True; end if;
    return (T1.R = T2.R);
  end "=";

  procedure SetZero(P: in out Fourier2) is
  begin
    P.T.R := Rad;
    SetZero(P.C);
    if not STrunc then SetZero(P.E); end if;
  end SetZero;

  procedure SetZero(T: in FType; F: in out Fourier2) is
  begin
    F.T := T;
    SetZero(F.C);
    if not STrunc then SetZero(F.E); end if;
  end SetZero;

  procedure SetZero(Template: in Fourier2; F: in out Fourier2) is
  begin
    F.T := Template.T;
    SetZero(F.C);
    if not STrunc then SetZero(F.E); end if;
  end SetZero;

  procedure SetZero(P: in ParPair; R: in RadPair; F: in out Fourier2) is
  begin
    F.T.P := P;
    F.T.R := R;
    SetZero(F.C);
    if not STrunc then SetZero(F.E); end if;
  end SetZero;

  procedure Copy(P1: in Fourier2; P2: in out Fourier2) is
  begin
    P2.T := P1.T;
    Copy(P1.C,P2.C);
    if not STrunc then Copy(P1.E,P2.E); end if;
  end Copy;

  procedure Swap(P1,P2: in out Fourier2) is
    T1: constant FType := P1.T;
  begin
    P1.T := P2.T;
    P2.T := T1;
    Swap(P1.C,P2.C);
    if not STrunc then Swap(P1.E,P2.E); end if;
  end Swap;

  procedure BallAt0(R: in Radius; P: in out Fourier2) is
  begin
    SetZero(P.C);
    if not STrunc then
      SetZero(P.E);
      P.E(0,0) := R;
    end if;
  end BallAt0;

  function BallAt0(R: Radius) return Fourier2 is
--- dangerous: P.T is not set
    P: Fourier2;
  begin
    BallAt0(R,P);
    return P;
  end BallAt0;
  
  procedure ToErr(P1,P2: in out Fourier2) is
    C1: FCoeff renames P1.C;
    C2: FCoeff renames P2.C;
    E1: FErr renames P1.E;
    E2: FErr renames P2.E;
    W1: constant Radius := NotExp(P1.T.R(1));
    W2: constant Radius := NotExp(P1.T.R(2));
    Fac1,Fac2: Flt;
    E: Scalar;
  begin
    P2.T := P1.T;
    SetZero(C2);
    if not STrunc then
      E2 := E1;
      SetZero(E1);
      Fac1 := One;
      for I in Freq1 loop
        Fac2 := Fac1;
        for J in Freq2 loop
          ModCenter(C1(I,J),E);
          E2(I,J) := E2(I,J)+Fac2*MaxNorm(E);
          Center(C1(I,J));
          Fac2 := W2*Fac2;
        end loop;
        Fac1 := W1*Fac1;
      end loop;
    end if;
  end ToErr;
  
  procedure Center(P: in out Fourier2) is
  begin
    if not STrunc then
      Center(P.C);
      SetZero(P.E);
    end if;
  end Center;

  procedure Center(P1: in Fourier2; P2: in out Fourier2) is
--- domain irrelevant now  
  begin
    Center(P1.C,P2.C);
    if not STrunc then SetZero(P2.E); end if;
    P2.T := P1.T;
  end Center;

  function Center(P: Fourier2) return Fourier2 is
    Q: Fourier2;
  begin
    Center(P,Q);
    return Q;
  end Center;

  procedure ModCenter(P: in out Fourier2) is
  begin
    ModCenter(P.C);
  end ModCenter;

  procedure ModCenter(P1: in Fourier2; P2: in out Fourier2) is
  begin
    ModCenter(P1.C,P2.C);
    if not STrunc then P2.E := P1.E; end if;
    P2.T := P1.T;
  end ModCenter;

  function ModCenter(P: Fourier2) return Fourier2 is
    Q: Fourier2;
  begin
    ModCenter(P,Q);
    return Q;
  end ModCenter;

  --- addition and multiplication etc.

  procedure Neg(P: in out Fourier2) is
  begin
    Neg(P.C);
  end Neg;

  procedure Neg(P1: in Fourier2; P2: in out Fourier2) is
  begin
    Neg(P1.C,P2.C);
    if not STrunc then P2.E := P1.E; end if;
    P2.T := P1.T;
  end Neg;

  function "-"(P: Fourier2) return Fourier2 is
    Q: Fourier2;
  begin
    Neg(P.C,Q.C);
    if not STrunc then Q.E := P.E; end if;
    Q.T := P.T;
    return Q;
  end "-";

  procedure Add(I: in Integer; P: in out Fourier2) is
  begin
    if P.T.P=(0,0) then
      Add(I,P.C(0,0));
    elsif I /= 0 then
      raise Parity_Error with "Fouriers2.Add error";
    end if;
  end Add;

  procedure Add(S: in Scalar; P: in out Fourier2) is
  begin
    if P.T.P=(0,0) then
      Add(S,P.C(0,0));
    elsif not IsZero(S) then
      raise Parity_Error with "Fouriers2.Add error";
    end if;
  end Add;

  procedure Add(P1: in Fourier2; P2: in out Fourier2) is
    T1: FType renames P1.T;
    T2: FType renames P2.T;
  begin
    if T1.P=T2.P then
      Add(P1.C,P2.C);
      if not STrunc then Add(P1.E,P2.E); end if;
      if T2.R(1)>T1.R(1) then T2.R(1) := T1.R(1); end if;
      if T2.R(2)>T1.R(2) then T2.R(2) := T1.R(2); end if;
    elsif IsZero(P2) then
      Copy(P1,P2);
    elsif not IsZero(P1) then
      raise Parity_Error with "Fouriers2.Add error";
    end if;
  end Add;

  procedure Sum(P1,P2: in Fourier2; P3: in out Fourier2) is
    T1: FType renames P1.T;
    T2: FType renames P2.T;
    T3: FType renames P3.T;
  begin
    if T1.P=T2.P then
      T3.P := T2.P;
      Sum(P1.C,P2.C,P3.C);
      if not STrunc then Sum(P1.E,P2.E,P3.E); end if;
      T3.R(1) := RMin(T1.R(1),T2.R(1));
      T3.R(2) := RMin(T1.R(2),T2.R(2));
    elsif IsZero(P2) then
      Copy(P1,P3);
    elsif IsZero(P1) then
      Copy(P2,P3);
    else
      raise Parity_Error with "Fouriers2.Sum error";
    end if;
  end Sum;

  function "+"(P1,P2: Fourier2) return Fourier2 is
    P3: Fourier2;
  begin
    Sum(P1,P2,P3);
    return P3;
  end "+";

  procedure Sub(P1: in Fourier2; P2: in out Fourier2) is
    T1: FType renames P1.T;
    T2: FType renames P2.T;
  begin
    if T1.P=T2.P then
      Sub(P1.C,P2.C);
      if not STrunc then Add(P1.E,P2.E); end if;
      if T2.R(1)>T1.R(1) then T2.R(1) := T1.R(1); end if;
      if T2.R(2)>T1.R(2) then T2.R(2) := T1.R(2); end if;
    elsif IsZero(P2) then
      Neg(P1,P2);
    elsif not IsZero(P1) then
      raise Parity_Error with "Fouriers2.Sub error";
    end if;
  end Sub;

  procedure Diff(P1,P2: in Fourier2; P3: in out Fourier2) is
    T1: FType renames P1.T;
    T2: FType renames P2.T;
    T3: FType renames P3.T;
  begin
    if T1.P=T2.P then
      T3.P := T2.P;
      Diff(P1.C,P2.C,P3.C);
      if not STrunc then Sum(P1.E,P2.E,P3.E); end if;
      T3.R(1) := RMin(T1.R(1),T2.R(1));
      T3.R(2) := RMin(T1.R(2),T2.R(2));
    elsif IsZero(P2) then
      Copy(P1,P3);
    elsif IsZero(P1) then
      Neg(P2,P3);
    else
      raise Parity_Error with "Fouriers2.Diff error";
    end if;
  end Diff;

  function "-"(P1,P2: Fourier2) return Fourier2 is
    P3: Fourier2;
  begin
    Diff(P1,P2,P3);
    return P3;
  end "-";

  procedure Mult(R: in Flt; P,Tmp: in out Fourier2) is
    STmp: Scalar renames Tmp.C(0,0);
  begin
    Mult(R,P.C,STmp);
    if not STrunc then Mult(abs(R),P.E); end if;
  end Mult;

  procedure Mult(R: in Flt; P: in out Fourier2) is
    STmp: Scalar;
  begin
    Mult(R,P.C,STmp);
    if not STrunc then Mult(abs(R),P.E); end if;
  end Mult;

  procedure Prod(R: in Flt; P1: in Fourier2; P2: in out Fourier2) is
  begin
    P2.T := P1.T;
    if R=Zero then
      SetZero(P2);
    else
      Prod(R,P1.C,P2.C);
      if not STrunc then Prod(abs(R),P1.E,P2.E); end if;
    end if;
  end Prod;

  function "*"(R: Flt; P: Fourier2) return Fourier2 is
    Q: Fourier2;
  begin
    Prod(R,P,Q);
    return Q;
  end "*";

  procedure AddProd(R: in Flt; P1: in Fourier2; P2: in out Fourier2; Tmp: in out Scalar) is
    T1: FType renames P1.T;
    T2: FType renames P2.T;
  begin
    if R /= Zero then
      if T1.P=T2.P then
        AddProd(R,P1.C,P2.C,Tmp);
        if not STrunc then AddProd(abs(R),P1.E,P2.E); end if;
        if T2.R(1)>T1.R(1) then T2.R(1) := T1.R(1); end if;
        if T2.R(2)>T1.R(2) then T2.R(2) := T1.R(2); end if;  -- P2 := P2 + R*P1
      elsif IsZero(P2) then
        Prod(R,P1,P2);
      elsif not IsZero(P1) then
        raise Parity_Error with "Fouriers2.AddProd error";
      end if;
    end if;
  end AddProd;

  procedure AddProd(R: in Flt; P1: in Fourier2; P2,Tmp: in out Fourier2) is
  begin
    AddProd(R,P1,P2,Tmp.C(0,0));
  end AddProd;

  procedure AddProd(R: in Flt; P1: in Fourier2; P2: in out Fourier2) is
    Tmp: Scalar;
  begin
    AddProd(R,P1,P2,Tmp);
  end AddProd;

  procedure Mult(Q: in Rational; P: in out Fourier2) is
  begin
    if LNum(Q)=0 then
      SetZero(P);
    else
      declare
        Tmp: Scalar;
      begin
        Mult(Flt(LNum(Q)),P.C,Tmp);
        Div(Flt(LDen(Q)),P.C,Tmp);
      end;
    end if;
  end Mult;

  procedure Mult(S: in Scalar; P: in out Fourier2) is
  begin
    Mult(S,P.C);
    if not STrunc then Mult(MaxNorm(S),P.E); end if;
  end Mult;

  procedure Prod(S: in Scalar; P1: in Fourier2; P2: in out Fourier2) is
  begin
    P2.T := P1.T;
    if IsZero(S) then
      SetZero(P2);
    else
      Prod(S,P1.C,P2.C);
      if not STrunc then Prod(MaxNorm(S),P1.E,P2.E); end if;
    end if;
  end Prod;

  function "*"(S: Scalar; P: Fourier2) return Fourier2 is
    Q: Fourier2;
  begin
    Prod(S,P,Q);
    return Q;
  end "*";

  procedure AddProd(S: in Scalar; P1: in Fourier2; P2: in out Fourier2) is
    T1: FType renames P1.T;
    T2: FType renames P2.T;
  begin
    if not IsZero(S) then
      if T1.P=T2.P then
        AddProd(S,P1.C,P2.C);
        if not STrunc then AddProd(MaxNorm(S),P1.E,P2.E); end if;
        if T2.R(1)>T1.R(1) then T2.R(1) := T1.R(1); end if;
        if T2.R(2)>T1.R(2) then T2.R(2) := T1.R(2); end if;
      elsif IsZero(P2) then
        Prod(S,P1,P2);
      elsif not IsZero(P1) then
        raise Parity_Error with "Fouriers2.AddProd error";
      end if;
    end if;
  end AddProd;

  procedure Mult(P1: in Fourier2; P2,Tmp: in out Fourier2) is
  begin
    Prod(P1,P2,Tmp);
    Copy(Tmp,P2);
  end Mult;

  procedure LMult(P1: in Fourier2; P2: in out Fourier2) is
    Tmp: LF_Pointer;
  begin
    Pool.Allocate(Tmp);
    Prod(P1,P2,Tmp.Data);
    Copy(Tmp.Data,P2);
    Pool.Recycle(Tmp);
  end LMult;

  procedure RMult(P1: in Fourier2; P2: in out Fourier2) is
    Tmp: LF_Pointer;
  begin
    Pool.Allocate(Tmp);
    Prod(P2,P1,Tmp.Data);
    Copy(Tmp.Data,P2);
    Pool.Recycle(Tmp);
  end RMult;

  --  2*cos(m)*cos(n) =  cos(m-n) + cos(m+n)
  --  2*sin(m)*sin(n) =  cos(m-n) - cos(m+n)
  --  2*cos(m)*sin(n) = -sin(m-n) + sin(m+n)
  --  2*sin(m)*cos(n) =  sin(m-n) + sin(m+n)

  function "*"(P1,P2: ParPair) return ParPair is
  begin
    return ((P1(1)+P2(1)) mod 2,(P1(2)+P2(2)) mod 2);
  end "*";

  procedure Signs(P1,P2: in Parity; Sum,DiffPlus,DiffMinus: out Integer) is
  begin
    Sum := 1;
    DiffPlus := 1;
    DiffMinus := 1;
    if P1=0 then
      if P2=1 then  DiffPlus := -1; end if;
    elsif P2=0 then
      DiffMinus := -1;
    else
      Sum := -1;
    end if;
  end Signs;

  procedure NumProd(F1,F2: in Fourier2; F3: in out Fourier2) is
    FC1: FCoeff renames F1.C;
    FC2: FCoeff renames F2.C;
    PC: Protected_Counter;

    procedure Prod_Loop(FC3: in out FCoeff; Xs,Xdp,Xdm,Ys,Ydp,Ydm: in Integer) is
      XsYs: constant Boolean := (Xs=Ys);
      M1,MS,MD,NS,ND,Xd,Yd: Integer;
      C12: Scalar;
    begin
      SetZero(FC3);
      loop
        PC.Next(M1);
        exit when M1>Deg1;
        for N1 in Freq2 loop
          declare
            C1: Scalar renames FC1(M1,N1);
          begin
            if not IsZero(C1) then
              for M2 in Freq1 loop
                MS := M1+M2;
                if M2<M1 then
                  MD := M1-M2;
                  Xd := Xdp;
                else
                  MD := M2-M1;
                  Xd := Xdm;
                end if;
                for N2 in Freq2 loop
                  declare
                    C2: Scalar renames FC2(M2,N2);
                  begin
                    if not IsZero(C2) then
                      NS := N1+N2;
                      if N2<N1 then
                        ND := N1-N2;
                        Yd := Ydp;
                      else
                        ND := N2-N1;
                        Yd := Ydm;
                      end if;
                      Prod(C1,C2,C12);
                      if Xd=Yd then
                        Add(C12,FC3(MD,ND));
                      else
                        Sub(C12,FC3(MD,ND));
                      end if;
                      if MS <= Deg1 then
                        if Xs=Yd then
                          Add(C12,FC3(MS,ND));
                        else
                          Sub(C12,FC3(MS,ND));
                        end if;
                        if NS <= Deg2 then
                          if XsYs then
                            Add(C12,FC3(MS,NS));
                          else
                            Sub(C12,FC3(MS,NS));
                          end if;
                        end if;
                      end if;
                      if NS <= Deg2 then
                        if Xd=Ys then
                          Add(C12,FC3(MD,NS));
                        else
                          Sub(C12,FC3(MD,NS));
                        end if;
                      end if;
                    end if;
                  end;
                end loop;
              end loop;
            end if;
          end;
        end loop;
      end loop;
    end Prod_Loop;

    protected type PC3_Type is procedure Add(C: in FCoeff); end PC3_Type;
    protected body PC3_Type is
      procedure Add (C: in FCoeff) is
      begin
        Add(C,F3.C);
      end Add;
    end PC3_Type;

    Xs,Xdp,Xdm,Ys,Ydp,Ydm: Integer;
    PC3: PC3_Type;

    procedure Prod_Loop is
      Tmp: LF_Pointer;
    begin
      Pool.Allocate(Tmp);
      Prod_Loop(Tmp.Data.C,Xs,Xdp,Xdm,Ys,Ydp,Ydm);
      PC3.Add(Tmp.Data.C);
      Pool.Recycle(Tmp);
    exception
      when others => Task_Error := True; raise;
    end Prod_Loop;

    task type Prod_Task_Type is end Prod_Task_Type;
    task body Prod_Task_Type is begin Prod_Loop; end Prod_Task_Type;

    NT: constant Natural := Reserve_Tasks(Deg1,FProd_Parallel);
    T1: FType renames F1.T;
    T2: FType renames F2.T;
    T3: FType renames F3.T;
  begin
    T3.P := T1.P*T2.P;
    T3.R := Min(T1.R,T2.R);
    Signs(T1.P(1),T2.P(1),Xs,Xdp,Xdm);
    Signs(T1.P(2),T2.P(2),Ys,Ydp,Ydm);
    PC.Reset(0);
    if NT=0 then
      Prod_Loop(F3.C,Xs,Xdp,Xdm,Ys,Ydp,Ydm);
    else
      SetZero(F3.C);
      declare
        Prod_Task: array(1 .. NT) of Prod_Task_Type;
      begin
        Prod_Loop;
      end;
      Free_Tasks(NT);
      if Task_Error then raise Sorry with "Exception in Fouriers2.NumProd"; end if;
    end if;
    Mult(SQuarter,F3);     --- factor 1/4
  end NumProd;

  procedure MaxNorms(C: in FCoeff; E: out Flt_Matrix) is
  begin
    for I in Freq1 loop
      for J in Freq2 loop E(I,J) := MaxNorm(C(I,J)); end loop;
    end loop;
  end MaxNorms;

  procedure Prod(F1,F2: in Fourier2; F3: in out Fourier2; ZC1,ZE1,ZC2,ZE2: in Boolean) is
    --- assuming STrunc is False
    CC: constant Boolean := not (ZC1 or ZC2);
    CE: constant Boolean := not (ZC1 or ZE2);
    EC: constant Boolean := not (ZE1 or ZC2);
    EE: constant Boolean := not (ZE1 or ZE2);
    PCC,PCE,PEC,PEE: Protected_Counter;

    procedure CCProd_Loop(F: in out Fourier2; Xs,Xdp,Xdm,Ys,Ydp,Ydm: in Integer; W1, W2: in Flt_Vector) is
      XsYs: constant Boolean := (Xs=Ys);
      FC1: FCoeff renames F1.C;
      FC2: FCoeff renames F2.C;
      FC3: FCoeff renames F.C;
      FE3: FErr renames F.E;
      M1,MS,MD,NS,ND,Xd,Yd: Integer;
      NC12: Flt;
      C12: Scalar;
    begin
      loop
        PCC.Next(M1);
        exit when M1>Deg1;
        for N1 in Freq2 loop
          declare
            C1: Scalar renames FC1(M1,N1);
          begin
            if not IsZero(C1) then
              for M2 in Freq1 loop
                MS := M1+M2;
                if M2<M1 then
                  MD := M1-M2;
                  Xd := Xdp;
                else
                  MD := M2-M1;
                  Xd := Xdm;
                end if;
                for N2 in Freq2 loop
                  declare
                    C2: Scalar renames FC2(M2,N2);
                  begin
                    if not IsZero(C2) then
                      NS := N1+N2;
                      if N2<N1 then
                        ND := N1-N2;
                        Yd := Ydp;
                      else
                        ND := N2-N1;
                        Yd := Ydm;
                      end if;
                      Prod(C1,C2,C12);
                      if Xd=Yd then
                        Add(C12,FC3(MD,ND));
                      else
                        Sub(C12,FC3(MD,ND));
                      end if;
                      NC12 := MaxNorm(C12);
                      if MS <= Deg1 then
                        if Xs=Yd then
                          Add(C12,FC3(MS,ND));
                        else
                          Sub(C12,FC3(MS,ND));
                        end if;
                        if NS <= Deg2 then
                          if XsYs then
                            Add(C12,FC3(MS,NS));
                          else
                            Sub(C12,FC3(MS,NS));
                          end if;
                        else
                          FE3(MS,NS) := FE3(MS,NS)+W1(MS)*W2(NS)*NC12;
                        end if;
                      else
                        FE3(MS,ND) := FE3(MS,ND)+W1(MS)*W2(ND)*NC12;
                        FE3(MS,NS) := FE3(MS,NS)+W1(MS)*W2(NS)*NC12;
                      end if;
                      if NS <= Deg2 then
                        if Xd=Ys then
                          Add(C12,FC3(MD,NS));
                        else
                          Sub(C12,FC3(MD,NS));
                        end if;
                      else
                        FE3(MD,NS) := FE3(MD,NS)+W1(MD)*W2(NS)*NC12;
                      end if;
                    end if;
                  end;
                end loop;
              end loop;
            end if;
          end;
        end loop;
      end loop;
    end CCProd_Loop;

    procedure CEProd_Loop(FC1: in FCoeff; FE2: in FErr; F: in out Fourier2; W1,W2: in Flt_Vector) is
      FE3: FErr renames F.E;
      M1,MS,MD,NS,ND: Integer;
      E1,E2,ED1,ED2,Fac: Flt;
    begin
      loop
        PCE.Next(M1);
        exit when M1>Deg1;
        for N1 in Freq2 loop
          E1 := MaxNorm(FC1(M1,N1));
          if E1 /= Zero then
            E1 := W1(M1)*W2(N1)*E1;
            for M2 in FreqE1 loop
              MS := IMin(M2+M1,DegE1);
              MD := IMax(M2-M1,0);
              Fac := W1(-IMin(M1,M2));
              ED1 := Fac*Fac*E1;
              for N2 in FreqE2 loop
                E2 := FE2(M2,N2);
                if E2 /= Zero then
                  NS := IMin(N2+N1,DegE2);
                  ND := IMax(N2-N1,0);
                  Fac := W2(-IMin(N1,N2));
                  ED2 := Fac*Fac*E2;
                  FE3(MS,NS) := FE3(MS,NS)+E1*E2;
                  FE3(MS,ND) := FE3(MS,ND)+E1*ED2;
                  FE3(MD,NS) := FE3(MD,NS)+ED1*E2;
                  FE3(MD,ND) := FE3(MD,ND)+ED1*ED2;
                end if;
              end loop;
            end loop;
          end if;
        end loop;
      end loop;
    end CEProd_Loop;

    procedure ECProd_Loop(FE2: in FErr; FC1: in FCoeff; F: in out Fourier2; W1,W2: in Flt_Vector) is
      FE3: FErr renames F.E;
      M1,MS,MD,NS,ND: Integer;
      E1,E2,ED1,ED2,Fac: Flt;
    begin
      loop
        PEC.Next(M1);
        exit when M1>Deg1;
        for N1 in Freq2 loop
          E1 := MaxNorm(FC1(M1,N1));
          if E1 /= Zero then
            E1 := W1(M1)*W2(N1)*E1;
            for M2 in FreqE1 loop
              MS := IMin(M2+M1,DegE1);
              MD := IMax(M2-M1,0);
              Fac := W1(-IMin(M1,M2));
              ED1 := Fac*Fac*E1;
              for N2 in FreqE2 loop
                E2 := FE2(M2,N2);
                if E2 /= Zero then
                  NS := IMin(N2+N1,DegE2);
                  ND := IMax(N2-N1,0);
                  Fac := W2(-IMin(N1,N2));
                  ED2 := Fac*Fac*E2;
                  FE3(MS,NS) := FE3(MS,NS)+E1*E2;
                  FE3(MS,ND) := FE3(MS,ND)+E1*ED2;
                  FE3(MD,NS) := FE3(MD,NS)+ED1*E2;
                  FE3(MD,ND) := FE3(MD,ND)+ED1*ED2;
                end if;
              end loop;
            end loop;
          end if;
        end loop;
      end loop;
    end ECProd_Loop;

    procedure EEProd_Loop(F: in out Fourier2; W1,W2: in Flt_Vector) is
      FE1: FErr renames F1.E;
      FE2: FErr renames F2.E;
      FE3: FErr renames F.E;
      M1,MS,NS: Integer;
      E1,E2,ED1,ED2,Fac: Flt;
    begin
      loop
        PEE.Next(M1);
        exit when M1>DegE1;
        for N1 in FreqE2 loop
          E1 := FE1(M1,N1);
          if E1 /= Zero then
            for M2 in FreqE1 loop
              MS := IMin(M2+M1,DegE1);
              Fac := W1(-IMin(M1,M2));
              ED1 := Fac*Fac*E1;
              for N2 in FreqE2 loop
                E2 := FE2(M2,N2);
                if E2 /= Zero then
                  NS := IMin(N2+N1,DegE2);
                  Fac := W2(-IMin(N1,N2));
                  ED2 := Fac*Fac*E2;
                  FE3(MS,NS) := FE3(MS,NS)+E1*E2;
                  FE3(MS,0) := FE3(MS,0)+E1*ED2;
                  FE3(0,NS) := FE3(0,NS)+ED1*E2;
                  FE3(0,0) := FE3(0,0)+ED1*ED2;
                end if;
              end loop;
            end loop;
          end if;
        end loop;
      end loop;
    end EEProd_Loop;

    protected type PF3_Type is procedure Add(F: in Fourier2); end PF3_Type;
    protected body PF3_Type is
      procedure Add(F: in Fourier2) is
      begin
        Add(F.C,F3.C);
        Add(F.E,F3.E);
      end Add;
    end PF3_Type;

    Xs,Xdp,Xdm,Ys,Ydp,Ydm: Integer;
    W1: Flt_Vector(-DegE1 .. DegE1);
    W2: Flt_Vector(-DegE2 .. DegE2);
    PF3: PF3_Type;

    procedure Prod_Loop is
      Tmp: LF_Pointer;
    begin
      Pool.Allocate(Tmp);
      SetZero(Tmp.Data);
      if CC then CCProd_Loop(Tmp.Data,Xs,Xdp,Xdm,Ys,Ydp,Ydm,W1,W2); end if;
      if CE then CEProd_Loop(F1.C,F2.E,Tmp.Data,W1,W2); end if;
      if EC then ECProd_Loop(F1.E,F2.C,Tmp.Data,W1,W2); end if;
      if EE then EEProd_Loop(Tmp.Data,W1,W2); end if;
      PF3.Add(Tmp.Data);
      Pool.Recycle(Tmp);
    exception
      when others => Task_Error := True; raise;
    end Prod_Loop;

    task type Prod_Task_Type is end Prod_Task_Type;
    task body Prod_Task_Type is begin Prod_Loop; end Prod_Task_Type;

    NT: constant Natural := Reserve_Tasks(Deg1,FProd_Parallel);
    T1: FType renames F1.T;
    T2: FType renames F2.T;
    T3: FType renames F3.T;
  begin
    T3.P := T1.P*T2.P;
    T3.R := Min(T1.R,T2.R);
    SetZero(F3.C);
    SetZero(F3.E);
    Signs(T1.P(1),T2.P(1),Xs,Xdp,Xdm);
    Signs(T1.P(2),T2.P(2),Ys,Ydp,Ydm);
    NormFacs(NotExp(T3.R(1)),W1);
    NormFacs(NotExp(T3.R(2)),W2);
    PCC.Reset(0);
    PCE.Reset(0);
    PEC.Reset(0);
    PEE.Reset(0);
    if NT=0 then
      Prod_Loop;
    else
      declare
        Prod_Task: array(1 .. NT) of Prod_Task_Type;
      begin
        Prod_Loop;
      end;
      Free_Tasks(NT);
      if Task_Error then raise Sorry with "Exception in Fouriers2.Prod"; end if;
    end if;
    Mult(SQuarter,F3);     --- factor 1/4
  end Prod;

  procedure Prod(P1,P2: in Fourier2; P3: in out Fourier2) is
    ZE1: constant Boolean := STrunc or else IsZero(P1.E);
    ZE2: constant Boolean := STrunc or else IsZero(P2.E);
    ZC1,ZC2: Boolean;
  begin
    ZC1 := IsZero(P1.C);
    if ZC1 and then ZE1 then SetZero(P3); return; end if;
    ZC2 := IsZero(P2.C);
    if ZC2 and then ZE2 then SetZero(P3); return; end if;
    if STrunc then
      NumProd(P1,P2,P3);
    else
      Prod(P1,P2,P3,ZC1,ZE1,ZC2,ZE2);
    end if;
  end Prod;

  function "*"(P1,P2: Fourier2) return Fourier2 is
    P3: Fourier2;
  begin
    Prod(P1,P2,P3);
    return P3;
  end "*";

  procedure AddProd(P1,P2: in Fourier2; P3: in out Fourier2) is
    Tmp: LF_Pointer;
  begin
    Pool.Allocate(Tmp);
    Prod(P1,P2,Tmp.Data);
    Add(Tmp.Data,P3);
    Pool.Recycle(Tmp);
  end AddProd;

  procedure Div(R: in Flt; P,Tmp: in out Fourier2) is
    E: FErr renames P.E;
    A: Flt renames Tmp.E(0,0);
    STmp: Scalar renames Tmp.C(0,0);
  begin
    Div(R,P.C,STmp);
    if not STrunc then
      A := abs(R);
      for I in FreqE1 loop
        for J in FreqE2 loop
          E(I,J) := E(I,J)/A;
        end loop;
      end loop;
    end if;
  end Div;

  procedure Div(R: in Flt; P: in out Fourier2) is
    E: FErr renames P.E;
    A: Flt;
    STmp: Scalar;
  begin
    Div(R,P.C,STmp);
    if not STrunc then
      A := abs(R);
      for I in FreqE1 loop
        for J in FreqE2 loop
          E(I,J) := E(I,J)/A;
        end loop;
      end loop;
    end if;
  end Div;

  procedure Quot(P1: in Fourier2; R: in Flt; P2: in out Fourier2) is
    E1: FErr renames P1.E;
    E2: FErr renames P2.E;
    A: Flt;
  begin
    Quot(P1.C,R,P2.C);
    if not STrunc then
      A := abs(R);
      for I in FreqE1 loop
        for J in FreqE2 loop
          E2(I,J) := E1(I,J)/A;
        end loop;
      end loop;
    end if;
    P2.T := P1.T;
  end Quot;

  function "/"(P: Fourier2; R: Flt) return Fourier2 is
    Q: Fourier2;
  begin
    Quot(P,R,Q);
    return Q;
  end "/";

  procedure MaxNorm(E: in FErr; N: out Radius; Count: out Natural) is
  begin
    N := Zero;
    Count := 0;
    if not STrunc then
      for I in FreqE1 loop
        for J in FreqE2 loop
          if E(I,J) /= Zero then
            N := N+E(I,J);
            Count := Count+1;
          end if;
        end loop;
      end loop;
    end if;
  end MaxNorm;

  function MaxNorm(E: FErr) return Radius is
    R: Flt := Zero;
  begin
    for M in FreqE1 loop
      for N in FreqE2 loop
        R := R+E(M,N);
      end loop;
    end loop;
    return R;
  end MaxNorm;
  
  function MaxNorm0(C: FCoeff) return Radius is
    R: Flt := Zero;
  begin
    for M in Freq1 loop
      for N in Freq2 loop
        R := R+MaxNorm(C(M,N));
      end loop;
    end loop;
    return R;
  end MaxNorm0;
  
  function MaxNorm(R: RadPair; C: FCoeff) return Radius is
    W1: constant Radius := NotExp(R(1));
    W2: constant Radius := NotExp(R(2));
    SM,S: Flt;
  begin
    S := Zero;
    for M in reverse Freq1 loop
      SM := Zero;
      for N in reverse Freq2 loop
        SM :=  W2*SM+MaxNorm(C(M,N));
      end loop;
      S := W1*S+SM;
    end loop;
    return S;
  end MaxNorm;

  function MaxNorm(R: RadPair; F: Fourier2) return Radius is
  begin
    if STrunc then return MaxNorm(R,F.C); end if;
    if F.T.R(1)<R(1) or else F.T.R(2)<R(2) then raise Domain_Error; end if;
    return MaxNorm(R,F.C)+MaxNorm(F.E);
  end MaxNorm;

  function MaxNorm(F: Fourier2) return Radius is
  begin
    if STrunc then return MaxNorm(F.T.R,F.C); end if;
    return MaxNorm(F.T.R,F.C)+MaxNorm(F.E);
  end MaxNorm;

  procedure Norm(R: in RadPair; C: in FCoeff; S: in out Scalar) is
  begin
    SetZero(S);
    if not IsZero(C) then
      declare
        W1: constant Scalar := NotExp(R(1));
        W2: constant Scalar := NotExp(R(2));
        SM,SN: Scalar;
      begin
        for M in reverse Freq1 loop
          SetZero(SM);
          for N in reverse Freq2 loop
            Norm(C(M,N),SN);
            Mult(W2,SM);
            Add(SN,SM);
          end loop;
          Mult(W1,S);
          Add(SM,S);
        end loop;
      end;
    end if;
  end Norm;

  procedure Norm(R: in RadPair; F: in Fourier2; S: in out Scalar) is
    Count: Natural;
    N: Radius;
  begin
    Norm(R,F.C,S);
    MaxNorm(F.E,N,Count);
    if Count=0 then return; end if;
    if F.T.R(1)<R(1) or else F.T.R(2)<R(2) then raise Domain_Error; end if;
    Add(DiskAt0(N),S);
  end Norm;

  function Norm(R: RadPair; F: Fourier2) return Scalar is
    S: Scalar;
  begin
    Norm(R,F,S);
    return S;
  end Norm;

  function Norm(F: Fourier2) return Scalar is
    S: Scalar;
  begin
    Norm(F.T.R,F,S);
    return S;
  end Norm;

  function NormOrRadius(F: Fourier2) return Scalar is
    Count: Natural;
    N: Radius;
    S: Scalar;
  begin
    Norm(F.T.R,F.C,S);
    MaxNorm(F.E,N,Count);
    if Count=0 then return S; end if;
    if Count=1 and then IsZero(S) then
      Assign(N,S); -- F is a ball and S is its radius
    else
      Add(DiskAt0(N),S);
    end if;
    return S;
  end NormOrRadius;

  procedure Sqr(P: in out Fourier2) is
    Tmp: LF_Pointer;
  begin
    Pool.Allocate(Tmp);
    Prod(P,P,Tmp.Data);
    Copy(Tmp.Data,P);
    Pool.Recycle(Tmp);
  end Sqr;

  function Sqr(P: Fourier2) return Fourier2 is
    Q: Fourier2;
  begin
    Prod(P,P,Q);
    return Q;
  end Sqr;

  --- conversion and i/o

  function Approx(P: Fourier2) return Flt is
  begin
    return Approx(P.C(0,0));
  end Approx;

  procedure Assign(I: in Integer; P: in out Fourier2) is
  begin
    P.T := ((0,0),Rad);
    SetZero(P.C);
    if not STrunc then SetZero(P.E); end if;
    Assign(I,P.C(0,0));
  end Assign;

  procedure Assign(Q: in Rational; P: in out Fourier2) is
  begin
    P.T := ((0,0),Rad);
    SetZero(P.C);
    if not STrunc then SetZero(P.E); end if;
    Assign(Q,P.C(0,0));
  end Assign;

  procedure Assign(R: in Flt; P: in out Fourier2) is
  begin
    P.T := ((0,0),Rad);
    SetZero(P.C);
    if not STrunc then SetZero(P.E); end if;
    Assign(R,P.C(0,0));
  end Assign;

  function Scal(I: Integer) return Fourier2 is
    P: Fourier2;
  begin
    P.T := ((0,0),Rad);
    SetZero(P.C);
    if not STrunc then SetZero(P.E); end if;
    Assign(I,P.C(0,0));
    return P;
  end Scal;

  function Scal(Q: Rational) return Fourier2 is
    P: Fourier2;
  begin
    P.T := ((0,0),Rad);
    SetZero(P.C);
    if not STrunc then SetZero(P.E); end if;
    Assign(Q,P.C(0,0));
    return P;
  end Scal;

  function Scal(R: Flt) return Fourier2 is
    P: Fourier2;
  begin
    P.T := ((0,0),Rad);
    SetZero(P.C);
    if not STrunc then SetZero(P.E); end if;
    Assign(R,P.C(0,0));
    return P;
  end Scal;

  function Scal(S: Scalar) return Fourier2 is
    P: Fourier2;
  begin
    P.T := ((0,0),Rad);
    SetZero(P.C);
    if not STrunc then SetZero(P.E); end if;
    Copy(S,P.C(0,0));
    return P;
  end Scal;

  procedure Enclose(R1,R2: in Flt; P: in out Fourier2) is
  begin
    P.T := ((0,0),Rad);
    SetZero(P.C);
    if not STrunc then SetZero(P.E); end if;
    Enclose(R1,R2,P.C(0,0));
  end Enclose;

  function Enclose(R1,R2: Flt) return Fourier2 is
    P: Fourier2;
  begin
    P.T := ((0,0),Rad);
    SetZero(P.C);
    if not STrunc then SetZero(P.E); end if;
    Enclose(R1,R2,P.C(0,0));
    return P;
  end Enclose;

  procedure Show1(S: in String; T: in FType) is
  begin
    if T.P(1)=0 then
      Show0(S & " cos*",False);
    else
      Show0(S & " sin*",False);
    end if;
    if T.P(2)=0 then
      Show2("cos  Rho",T.R(1),T.R(2));
    else
      Show2("sin  Rho",T.R(1),T.R(2));
    end if;
    New_Line;
  end Show1;

  procedure Show1(S: in String; F: in Fourier2) is
    W: constant Positive := 4;
  begin
    if IsZero(F) then
      Put(S & "  Zero_Fourier2");
    else
      Show1(S,F.T);
      Show1(S,F.C);
      if not STrunc then
        New_Line;
        for M in FreqE1 loop
          for N in FreqE2 loop
            if F.E(M,N) /= Zero then
              Show1(S & " Err " & Strng(M,W) & Strng(N,W),F.E(M,N));
            end if;
          end loop;
        end loop;
      end if;
    end if;
    New_Line;
  end Show1;

  procedure Show2(S: in String; F1,F2: in Fourier2) is
    W: constant Positive := 4;
  begin
    if IsZero(F1) and then IsZero(F2) then
      Put(S & "  Zero_Fouriers2");
    else
      Show1(S & "1st  ",F1.T);
      Show1(S & "2nd  ",F2.T);
      Show2(S,F1.C,F2.C);
      if not STrunc then
        New_Line;
        for M in FreqE1 loop
          for N in FreqE2 loop
            if (F1.E(M,N) /= Zero) or else (F2.E(M,N) /= Zero) then
              Show2(S & " Err " & Strng(M,W) & Strng(N,W),F1.E(M,N),F2.E(M,N));
            end if;
          end loop;
        end loop;
      end if;
    end if;
    New_Line;
  end Show2;

  function EffDeg1(C: FCoeff) return Integer is
  begin
    for M in reverse Freq1 loop
      for N in Freq2 loop
        if C(M,N) /= SZero then return M; end if;
      end loop;
    end loop;
    return -1;
  end EffDeg1;

  function EffDeg2(C: FCoeff) return Integer is
  begin
    for N in reverse Freq2 loop
      for M in Freq1 loop
        if C(M,N) /= SZero then return N; end if;
      end loop;
    end loop;
    return -1;
  end EffDeg2;

  procedure EffDegs0(C: in FCoeff; D1,D2: out Integer) is
  begin
    D1 := IMax(EffDeg1(C),0);
    D2 := IMax(EffDeg2(C),0);
  end EffDegs0;

  function EffDeg1(E: FErr) return Integer is
  begin
    for M in reverse Freq1 loop
      for N in Freq2 loop
        if E(M,N) /= Zero then return M; end if;
      end loop;
    end loop;
    return -1;
  end EffDeg1;

  function EffDeg2(E: FErr) return Integer is
  begin
    for N in reverse Freq2 loop
      for M in Freq1 loop
        if E(M,N) /= Zero then return N; end if;
      end loop;
    end loop;
    return -1;
  end EffDeg2;

  procedure EffDegs0(E: in FErr; D1,D2: out Integer) is
  begin
    D1 := IMax(0,EffDeg1(E));
    D2 := IMax(0,EffDeg2(E));
  end EffDegs0;

  procedure Put(F: in File_Type; H: in Fourier2; Decimal: in Boolean := False) is
    D1,D2: Integer;
  begin
    Put(F,H.T.P(1));
    Put(F,H.T.P(2));
    Put(F,H.T.R(1),Decimal);
    Put(F,H.T.R(2),Decimal);
    EffDegs0(H.C,D1,D2);
    Put(F,D1);
    Put(F,D2);
    for M in 0 .. D1 loop
      for N in 0 .. D2 loop
        Put(F,H.C(M,N),Decimal);
      end loop;
    end loop;
    EffDegs0(H.E,D1,D2);
    Put(F,D1);
    Put(F,D2);
    for M in 0 .. D1 loop
      for N in 0 .. D2 loop
        Put(F,H.E(M,N),Decimal);
      end loop;
    end loop;
  end Put;

  procedure Get(F: in File_Type; H: in out Fourier2; Decimal: in Boolean := False) is
    HE: Ferr renames H.E;
    D1,D2: Integer := 0;
    M0,N0: Natural;
    E: Radius := Zero;
    C: Scalar;
  begin
    SetZero(H);
    Get(F,H.T.P(1));
    Get(F,H.T.P(2));
    Get(F,H.T.R(1),Decimal);
    Get(F,H.T.R(2),Decimal);
    Get(F,D1);
    Get(F,D2);
    for M in 0 .. D1 loop
      for N in 0 .. D2 loop
        Get(F,C,Decimal);
        AddCoeff(M,N,C,H);
      end loop;
    end loop;
    Get(F,D1);
    Get(F,D2);
    for M in 0 .. D1 loop
      M0 := IMin(M,DegE1);
      for N in 0 .. D2 loop
        N0 := IMin(N,DegE2);
        Get(F,E,Decimal);
        HE(M0,N0) := HE(M0,N0)+E;
      end loop;
    end loop;
  end Get;

  procedure Write(Name: in String; H: in Fourier2; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    if Verbosity>1 then Put_Line("Writing " & Name); end if;
    Create(F,Out_File,Name);
    Put(F,H,Decimal);
    Close(F);
  end Write;

  procedure Read(Name: in String; H: in out Fourier2; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    if Verbosity>1 then Put_Line("Reading " & Name); end if;
    Open(F,In_File,Name);
    Get(F,H,Decimal);
    Close(F);
  end Read;
  
  ------------------------
  
  procedure Show_Param_F2(N: in String := "") is
  begin
    if STrunc then
      Show0(N & "Scalar type is Numeric");
    else
      Show0(N & "Scalar type is Ball");
    end if;
    if FProd_Parallel then
      Show0(N & "FProd_Parallel is True");
    else
      Show0(N & "FProd_Parallel is False");
    end if;
    Show1(N & "Precision ",Get_Precision(SOne/Three));
    Show2(N & "Degs =",Deg1,Deg2);
  end Show_Param_F2;
  
  procedure NumRead(Name: in String; H: in out Fourier2; Decimal: in Boolean := False) is
    C: FCoeff renames H.C;
  begin
    Read(Name,H,Decimal);
    if not STrunc then
      for I in C'Range(1) loop
        for J in C'Range(2) loop
          Center(C(I,J));
        end loop;
      end loop;
      SetZero(H.E);
    end if;
  end NumRead;

  procedure Plot(G: in Fourier2; L: Positive := 50) is
    use Flt_IO;
    F: File_Type;
    P: constant Flt := Approx(Pi);
    X,Y,Z: Scalar;
  begin
    Create(F,Out_File,"out.dat");
    for M in 0 .. L loop
      for N in 0 .. L loop
        X := P*Scal(M)/Scal(L);
        Y := P*Scal(N)/Scal(L);
        Z := Val(X,Y,G);
        Put(F,Approx(X),3,3,3);
        Put(F,Approx(Y),3,3,3);
        Put(F,Approx(Z),3,3,3);
        New_Line(F);
      end loop;
    end loop;
    Close(F);
  end Plot;

  function GuessRho1(F: Fourier2) return Flt is
    use Flt_EF;
    C: FCoeff renames F.C;
    D1: constant Integer := EffDeg1(C);
    M0: constant Integer := (2*D1)/3;
    R,FM,T: Flt := Zero;
  begin
    for M in M0 .. D1 loop
      FM := Flt(M);
      for N in 0 .. Deg2 loop
        T := MaxNorm(C(M,N));
        if T /= Zero then
          R := RMax(Exp(Log(T)/FM),R);
        end if;
      end loop;
    end loop;
    return One/R-One;
  end GuessRho1;

  function GuessRho2(F: Fourier2) return Flt is
    use Flt_EF;
    C: FCoeff renames F.C;
    D2: constant Integer := EffDeg2(C);
    N0: constant Integer := (2*D2)/3;
    R,FN,T: Flt := Zero;
  begin
    for N in N0 .. D2 loop
      FN := Flt(N);
      for M in 0 .. Deg1 loop
        T := MaxNorm(C(M,N));
        if T /= Zero then
          R := RMax(Exp(Log(T)/FN),R);
        end if;
      end loop;
    end loop;
    return One/R-One;
  end GuessRho2;

  function GuessRho(F: Fourier2) return RadPair is
  begin
    return (GuessRho1(F),GuessRho2(F));
  end GuessRho;

  procedure GuessRho(F: in Fourier2) is
    T: FType;
  begin
    T.P := F.T.P;
    T.R := GuessRho(F);
    Show1("guessed domain:",T);
  end GuessRho;

  --- misc

  function Get_Precision(P: Fourier2) return Positive is
  begin
    return Get_Precision(P.C(0,0));
  end Get_Precision;

  procedure Proper_Rounding(Dummy: in Fourier2) is
    pragma Unreferenced (Dummy);
  begin
    Proper_Rounding(SZero);
  end Proper_Rounding;

  procedure Free_Cache_F2 is
  begin
    Pool.Free_All;
    Free_Cache(SZero);
  end Free_Cache_F2;
  
  procedure Free_Cache(Dummy: in Fourier2) is
    pragma Unreferenced (Dummy);
  begin
    Pool.Free_All;
    Free_Cache(SZero);
  end Free_Cache;
  
  --------------- other standard procedures

  function Power(R: Radius; I: Natural) return Radius is
  begin
    if I>1 then
      declare
        J: Integer := I;
        X: Flt := R;
        Y: Flt := One;
      begin
        while J>0 loop
          if (J rem 2)=1 then
            Y := X*Y;
          end if;
          X := X*X;
          J := J/2;
        end loop;
        return X*Y;
      end;
    elsif I=1 then
      return R;
    else
      return One;
    end if;
  end Power;

  procedure NormFacs(R: in Radius; W: in out Flt_Vector) is
    M,N: Integer;
  begin
    W(0) := NotExp(R);
    if W'Last>0 then
      M := 1;
      W(1) := R;
      for K in 2 .. W'Last loop
        N := K-M;
        W(K) := W(M)*W(N);
        if N=M then M := K; end if;
      end loop;
    end if;
    if W'First<0 then
      M := -1;
      W(-1) := NotExp(-R);
      for K in reverse W'First .. -2 loop
        N := K-M;
        W(K) := W(M)*W(N);
        if N=M then M := K; end if;
      end loop;
    end if;
  end NormFacs;

  function NumCoeffs(D1,D2: Natural; JustOdd: Boolean) return Positive is
    --- number of coefficients up to degrees D1,D2
  begin
    if JustOdd then
      if ((D1*D2) mod 2)=0 then
        raise Parity_Error with "Fouriers2.NumCoeffs: expect odd degrees";
      end if;
      return (D1/2+1)*(D2/2+1);
    else
      return (D1+1)*(D2+1);
    end if;
  end NumCoeffs;

  procedure Coeff(M,N: in Natural; F: in Fourier2; S: in out Scalar) is
    --- too slow for use in loops
  begin
    if STrunc then
      if (M <= Deg1) and then (N <= Deg2) then
        Copy(F.C(M,N),S);
      else
        SetZero(S);
      end if;
    else
      declare
        M0: constant Natural := IMin(M,DegE1);
        N0: constant Natural := IMin(N,DegE2);
        R: RadPair renames F.T.R;
        FE: FErr renames F.E;
        E: Flt := Zero;
      begin
        for M1 in 0 .. M0 loop
          for N1 in 0 .. N0 loop E := E+FE(M1,N1); end loop;
        end loop;
        DiskAt0(Power(NotExp(-R(1)),M)*Power(NotExp(-R(2)),N)*E,S);
      end;
      if (M <= Deg1) and then (N <= Deg2) then
        Add(F.C(M,N),S);
      end if;
    end if;
  end Coeff;

  function Coeff(M,N: Natural; F: Fourier2) return Scalar is
    S: Scalar;
  begin
    Coeff(M,N,F,S);
    return S;
  end Coeff;

  procedure AddCoeff(M,N: in Natural; S: in Scalar; F: in out Fourier2) is
    --- too slow for use in loops
  begin
    if (M <= Deg1) and then (N <= Deg2) then
      Add(S,F.C(M,N));
    elsif not (STrunc or else IsZero(S)) then
      declare
        M0: constant Natural := IMin(M,DegE1);
        N0: constant Natural := IMin(N,DegE2);
        R: RadPair renames F.T.R;
      begin
        F.E(M0,N0) := F.E(M0,N0)+Power(NotExp(R(1)),M)*Power(NotExp(R(2)),N)*MaxNorm(S);
      end;
    end if;
  end AddCoeff;

  procedure AddBall(M,N: in Natural; R: in Flt; F: in out Fourier2) is
    M0: constant Natural := IMin(M,DegE1);
    N0: constant Natural := IMin(N,DegE2);
  begin
    if not STrunc then
      F.E(M0,N0) := F.E(M0,N0)+abs(R);
    end if;
  end AddBall;

  procedure AddBall(M,N: in Natural; S: in Scalar; F: in out Fourier2) is
    M0: constant Natural := IMin(M,DegE1);
    N0: constant Natural := IMin(N,DegE2);
  begin
    if not STrunc then
      F.E(M0,N0) := F.E(M0,N0)+MaxNorm(S);
    end if;
  end AddBall;

  procedure AddBall(R: in Flt; F: in out Fourier2) is
  begin
    if not STrunc then
      F.E(0,0) := F.E(0,0)+abs(R);
    end if;
  end AddBall;

  procedure AddBall(S: in Scalar; F: in out Fourier2) is
  begin
    if not STrunc then
      F.E(0,0) := F.E(0,0)+MaxNorm(S);
    end if;
  end AddBall;

  function Val(X,Y: Scalar; F: Fourier2) return Scalar is
    P1: constant Parity := F.T.P(1);
    P2: constant Parity := F.T.P(2);
    C: FCoeff renames F.C;
    S,SX,V,Tmp: Scalar;
  begin
    if STrunc then
      SetZero(V);
    else
      DiskAt0(MaxNorm(F.E),V);
    end if;
    for M in Freq1 loop
      Prod(Flt(M),X,Tmp);
      if P1=0 then Cos(Tmp,SX); else Sin(Tmp,SX); end if;
      for N in Freq2 loop
        Prod(Flt(N),Y,Tmp);
        if P2=0 then Cos(Tmp,S); else Sin(Tmp,S); end if;
        Mult(SX,S);
        Mult(C(M,N),S);
        Add(S,V);
      end loop;
    end loop;
    return V;
  end Val;
  
  procedure CleanOdd1(F: in out Fourier2) is
  begin
    if F.T.P(1)=1 then
      declare
        FC: FCoeff renames F.C;
        FE: FErr renames F.E;
      begin
        for K in Freq2 loop SetZero(FC(0,K)); end loop;
        if not STrunc then
          for K in FreqE2 loop
            FE(1,K) := FE(1,K)+FE(0,K);
            FE(0,K) := Zero;
          end loop;
        end if;
      end;
    end if;
  end CleanOdd1;
  
  procedure CleanOdd2(F: in out Fourier2) is
  begin
    if F.T.P(2)=1 then
      declare
        FC: FCoeff renames F.C;
        FE: FErr renames F.E;
      begin
        for N in Freq1 loop SetZero(FC(N,0)); end loop;
        if not STrunc then
          for N in FreqE1 loop
            FE(N,1) := FE(N,1)+FE(N,0);
            FE(N,0) := Zero;
          end loop;
        end if;
      end;
    end if;
  end CleanOdd2;
    
  procedure CleanOdd(F: in out Fourier2) is
  begin
    CleanOdd1(F);
    CleanOdd2(F);
  end CleanOdd;
  
  procedure ProjFreq1(P: in Parity; F: in out Fourier2) is
--- project onto even or odd (parity P) frequencies in 1st variable
    FC: FCoeff renames F.C;
    FE: FErr renames F.E;
    N: Integer := 1-P;
  begin
    if P=0 then CleanOdd1(F); end if;
    while N <= Deg1 loop
      for K in Freq2 loop SetZero(FC(N,K)); end loop;
      N := N+2;
    end loop;
    if STrunc then return; end if;
    N := 1-P;
    while N<DegE1 loop
      for K in FreqE2 loop
        FE(N+1,K) := FE(N+1,K)+FE(N,K);
        FE(N,K) := Zero;
      end loop;
      N := N+2;
    end loop;
  end ProjFreq1;
  
  procedure ProjEvenFreq1(F: in out Fourier2) is
  begin
    ProjFreq1(0,F);
  end ProjEvenFreq1;
  
  procedure ProjOddFreq1(F: in out Fourier2) is
  begin
    ProjFreq1(1,F);
  end ProjOddFreq1;
  
  procedure ProjFreq2(P: in Parity; F: in out Fourier2) is
--- project onto even or odd (parity P) frequencies in 2nd variable
    FC: FCoeff renames F.C;
    FE: FErr renames F.E;
    K: Integer := 1-P;
  begin
    if P=0 then CleanOdd2(F); end if;
    while K <= Deg2 loop
      for N in Freq1 loop SetZero(FC(N,K)); end loop;
      K := K+2;
    end loop;
    if STrunc then return; end if;
    K := 1-P;
    while K<DegE2 loop
      for N in FreqE1 loop
        FE(N,K+1) := FE(N,K+1)+FE(N,K);
        FE(N,K) := Zero;
      end loop;
      K := K+2;
    end loop;
  end ProjFreq2;
  
  procedure ProjEvenFreq2(F: in out Fourier2) is
  begin
    ProjFreq2(0,F);
  end ProjEvenFreq2;
  
  procedure ProjOddFreq2(F: in out Fourier2) is
  begin
    ProjFreq2(1,F);
  end ProjOddFreq2;
  
  procedure Separate1(E: in out FErr; D1,D2: in Natural) is
--- the new E(0..D1,0..D2) are independent of the first variable
    N1: Integer;  
  begin
    for N in 0 .. D1 loop
      N1 := N+1;
      for K in 0 .. D2 loop
        E(N1,K) := E(N1,K)+E(N,K);
      end loop;
    end loop;
  end Separate1;
  
  procedure Separate1(F: in out Fourier2; D: in Natural) is
--- the new F.E(0..D,*) are independent of the first variable
  begin
    if not STrunc then Separate1(F.E,D,DegE2); end if;
  end Separate1;
  
  procedure Separate2(E: in out FErr; D1,D2: in Natural) is
--- the new E(0..D1,0..D2) are independent of the second variable
    K1: Integer;  
  begin
    for K in 0 .. D2 loop
      K1 := K+1;
      for N in 0 .. D1 loop
        E(N,K1) := E(N,K1)+E(N,K);
      end loop;
    end loop;
  end Separate2;
  
  procedure Separate2(F: in out Fourier2; D: in Natural) is
--- the new F.E(*,0..D) are independent of the second variable
  begin
    if not STrunc then Separate2(F.E,DegE1,D); end if;
  end Separate2;
  
  procedure SetRho0(F: in out Fourier2) is
    R: RadPair renames F.T.R;
  begin
    if Strunc or else R=(Zero,Zero) then return; end if;
    declare
      W1: constant Radius := NotExp(-R(1));
      W2: constant Radius := NotExp(-R(2));
      Fac1,Fac2: Flt;
      FE: FErr renames F.E;
    begin
      Fac1 := One;
      for N in Freq1 loop
        Fac2 := Fac1;
        for K in Freq2 loop
          FE(N,K) := FE(N,K)*Fac2;
          Fac2 := W2*Fac2;
        end loop;
        Fac1 := W1*Fac1;
      end loop;
    end;
    R := (Zero,Zero);
  end SetRho0;
  
end Fouriers2;
