with Ada.Text_IO, Logicals, Scalars, Globals, Ints, Rationals, Flts, Flt_Matrices, Flt_Vectors;
use Ada.Text_IO, Logicals, Scalars, Globals, Ints, Rationals, Flts;
with Recycling, Matrices, Vectors, Matrices.More, Vectors.More;

pragma Elaborate_All (Ada.Text_IO,Logicals,Scalars,Globals,Ints,Rationals,Flts, Flt_Matrices,Flt_Vectors);
pragma Elaborate_All (Recycling,Matrices,Vectors,Matrices.More,Vectors.More);

generic

  Deg1,Deg2: in Positive;

-------------------- begin_include scalars Scalar 

  type Scalar is private;

--- basic
  with function Info(Dummy: Scalar) return Scalar_Info is <>;
  with function IsSharp(S: Scalar) return Boolean is <>;
  with function IsNumbers(S: Scalar) return Boolean is <>;
  with function IsZero(S: Scalar) return Boolean is <>;
  with function "="(S1,S2: Scalar) return Boolean is <>;
  with procedure SetZero(S: in out Scalar) is <>;
  with procedure Copy(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Swap(S1,S2: in out Scalar) is <>;

--- sets
  with function Center0(S: Scalar) return Boolean is <>;
  with function Contains0(S: Scalar) return Logical is <>;
  with function Contains(S1,S2: Scalar) return Logical is <>; -- S1 constains S2
  with procedure BallAt0(R: in Flt; S: in out Scalar) is <>;
  with function BallAt0(R: Flt) return Scalar is <>;
  with procedure DiskAt0(R: in Flt; S: in out Scalar) is <>;
  with function DiskAt0(R: Flt) return Scalar is <>;
  with procedure ToErr(S: in out Scalar) is <>;
  with procedure ToErr(S1: in Scalar; S2: in out Scalar) is <>;
  with function ToErr(S: Scalar) return Scalar is <>;
  with procedure Center(S: in out Scalar) is <>;
  with procedure Center(S1: in Scalar; S2: in out Scalar) is <>;
  with function Center(S: Scalar) return Scalar is <>;
  with procedure CenterDisk(S: in out Scalar) is <>;
  with procedure CenterDisk(S1: in Scalar; S2: in out Scalar) is <>;
  with function CenterDisk(S: Scalar) return Scalar is <>;
  with procedure ModCenter(S: in out Scalar) is <>;
  with procedure ModCenter(S1: in Scalar; S2: in out Scalar) is <>;
  with function ModCenter(S: Scalar) return Scalar is <>;
  with procedure ModNumbers(S: in out Scalar) is <>;
  with procedure ModNumbers(S1: in Scalar; S2: in out Scalar) is <>;
  with function ModNumbers(S: Scalar) return Scalar is <>;
  with procedure ErrMult(R: in Radius; S: in out Scalar) is <>;
  with procedure Union(S1: in Scalar; S2: in out Scalar) is <>;
  with function Union(S1,S2: Scalar) return Scalar is <>;
  with procedure Intersection(S1: in Scalar; S2: in out Scalar; Empty: out Logical) is <>;

--- order
  with function Sign(S: Scalar) return Integer is <>;
  with function Compare(S1,S2: Scalar) return Integer is <>;
  with function "<"(S1,S2: Scalar) return Boolean is <>;
  with function "<="(S1,S2: Scalar) return Boolean is <>;
  with function ">"(S1,S2: Scalar) return Boolean is <>;
  with function ">="(S1,S2: Scalar) return Boolean is <>;
  with procedure Min(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Min(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function Min(S1,S2: Scalar) return Scalar is <>;
  with procedure Max(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Max(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function Max(S1,S2: Scalar) return Scalar is <>;
  with function Inf(S: Scalar) return Flt is <>; -- inf of real part
  with function Sup(S: Scalar) return Flt is <>; -- sup of real part

--- arithmetic
  with procedure Neg(S: in out Scalar) is <>;
  with procedure Neg(S1: in Scalar; S2: in out Scalar) is <>;
  with function "-"(S: Scalar) return Scalar is <>;
  with procedure Add(I: in Integer; S: in out Scalar) is <>;
  with procedure Add(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Sum(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "+"(S1,S2: Scalar) return Scalar is <>;
  with procedure Sub(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Diff(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "-"(S1,S2: Scalar) return Scalar is <>;
  with procedure Mult(R: in Flt; S,Tmp: in out Scalar) is <>;
  with procedure Mult(R: in Flt; S: in out Scalar) is <>;
  with procedure Prod(R: in Flt; S1: in Scalar; S2: in out Scalar) is <>;
  with function "*"(R: Flt; S: Scalar) return Scalar is <>;
  with procedure AddProd(R: in Flt; S1: in Scalar; S2,Tmp: in out Scalar) is <>;
  with procedure AddProd(R: in Flt; S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Mult(Q: in Rational; S: in out Scalar) is <>;
  with procedure Mult(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Prod(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "*"(S1,S2: Scalar) return Scalar is <>;
  with procedure AddProd(S1,S2: in Scalar; S3,Tmp: in out Scalar) is <>;
  with procedure AddProd(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure SubProd(S1,S2: in Scalar; S3,Tmp: in out Scalar) is <>;
  with procedure Div(R: in Flt; S,Tmp: in out Scalar) is <>;
  with procedure Div(R: in Flt; S: in out Scalar) is <>;
  with procedure Quot(S1: in Scalar; R: in Flt; S2: in out Scalar) is <>;
  with function "/"(S: Scalar; R: Flt) return Scalar is <>;
  with procedure Div(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Quot(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "/"(S1,S2: Scalar) return Scalar is <>;
  with procedure Inv(S1: in Scalar; S2: in out Scalar) is <>;
  with function Inv(S: Scalar) return Scalar is <>;
  with function "**"(S: Scalar; I: Integer) return Scalar is <>;

--- fun
  with function IsReal(S: Scalar) return Boolean is <>;
  with procedure Adjoint(S: in out Scalar) is <>;
  with procedure Adjoint(S1: in Scalar; S2: in out Scalar) is <>;
  with function Adjoint(S: Scalar) return Scalar is <>;
  with procedure Real_Part(S: in out Scalar) is <>;
  with procedure Imag_Part(S: in out Scalar) is <>;
  with procedure Eval0(S: in out Scalar) is <>;
  with function "abs"(S: Scalar) return Scalar is <>;
  with procedure Norm(S1: in Scalar; S2: in out Scalar) is <>;
  with function Norm(S: Scalar) return Scalar is <>;
  with function MaxNorm(S: Scalar) return Radius is <>;
  with procedure Sqr(S: in out Scalar) is <>;
  with function Sqr(S: Scalar) return Scalar is <>;
  with procedure Sqrt(S1: in Scalar; S2: in out Scalar) is <>;
  with function Sqrt(S: Scalar) return Scalar is <>;
  with procedure Root(K: in Positive; S1: in Scalar; S2: in out Scalar) is <>;
  with function Root(K: Positive; S: Scalar) return Scalar is <>;
  with procedure Exp(S1: in Scalar; S2: in out Scalar) is <>;
  with function Exp(S: Scalar) return Scalar is <>;
  with procedure ArcCos(S1: in Scalar; S2: in out Scalar) is <>;
  with function ArcCos(S: Scalar) return Scalar is <>;
  with procedure ArcSin(S1: in Scalar; S2: in out Scalar) is <>;
  with function ArcSin(S: Scalar) return Scalar is <>;
  with procedure Cos(S1: in Scalar; S2: in out Scalar) is <>;
  with function Cos(S: Scalar) return Scalar is <>;
  with procedure Sin(S1: in Scalar; S2: in out Scalar) is <>;
  with function Sin(S: Scalar) return Scalar is <>;
  with procedure Log(S1: in Scalar; S2: in out Scalar) is <>;
  with function Log(S: Scalar) return Scalar is <>;
  with function Pi return Scalar is <>;
  with procedure Simple_Random(S: in out Scalar) is <>;
  with procedure QPower(Q: in Rational; S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Roots2(B,C: in Scalar; U1,U2,V: in out Scalar) is <>;
  with procedure Roots3(B,C,D: in Scalar; U0,U1,U2,V: in out Scalar) is <>;

--- conversion and i/o
  with function Approx(S: Scalar) return Flt is <>;
  with procedure Assign(K: in Integer; S: in out Scalar) is <>;
  with procedure Assign(Q: in Rational; S: in out Scalar) is <>;
  with procedure Assign(R: in Flt; S: in out Scalar) is <>;
  with function Scal(I: Integer) return Scalar is <>;
  with function Scal(Q: Rational) return Scalar is <>;
  with function Scal(R: Flt) return Scalar is <>;
  with procedure Enclose(R1,R2: in Flt; S: in out Scalar) is <>;
  with function Enclose(R1,R2: Flt) return Scalar is <>;
  with procedure Show1(N: in String; S: in Scalar; Whatever: in Boolean := True) is <>;
  with procedure Show2(N: in String; S1,S2: in Scalar; Whatever: in Boolean := True) is <>;
  with procedure Put(F: in File_Type; S: in Scalar; Decimal: in Boolean := False) is <>;
  with procedure Get(F: in File_Type; S: in out Scalar; Decimal: in Boolean := False) is <>;

--- misc
  with function Get_Precision(R: Scalar) return Positive is <>;
  with procedure Proper_Rounding(Dummy: in Scalar) is <>;
  with procedure Free_Cache(S: in Scalar) is <>;
-------------------- end_include scalars 

  with package SM is new Matrices (Scalar => Scalar);
  with package SV is new Vectors (Component => Scalar);

package Fouriers2 is  -- Fourier series in 2d with non-constant coefficients

  package SV_More is new SV.More;
  package SM_More is new SM.More;

  DegE1: constant Positive := 2*Deg1;
  DegE2: constant Positive := 2*Deg2;

  subtype Freq1  is Integer range 0 .. Deg1;
  subtype Freq2  is Integer range 0 .. Deg2;
  subtype FreqE1 is Integer range 0 .. DegE1;
  subtype FreqE2 is Integer range 0 .. DegE2;

  subtype FLt_Matrix is Flt_Matrices.Matrix;
  subtype FLt_Vector is FLt_Vectors.Vector;
  subtype FErr is  Flt_Matrix(FreqE1,FreqE2);
  subtype FCoeff is SM.Matrix(Freq1,Freq2);
  
  type FType is
    record
      P: ParPair;  --- even (0) or odd (1)
      R: RadPair;  --- domain parameters
    end record;

  type Fourier2 is
    record
      T: FType;
      C: FCoeff;
      E: FErr;   --- assumes Flt ops are rounded up
    end record;

  Rad: RadPair := (Two,Two);        -- default domain
  FProd_Parallel: Boolean := False;
  
--------------- standard Scalar procedures

--- basic
  function Info(Dummy: Fourier2) return Scalar_Info;
  function IsPoly(P: Fourier2) return Boolean;
  function IsSharp(P: Fourier2) return Boolean;         --- polynomial with sharp coeffs
  function IsZero(P: Fourier2) return Boolean;
  function "="(P1,P2: Fourier2) return Boolean;
  procedure SetZero(P: in out Fourier2);
  procedure SetZero(T: in FType; F: in out Fourier2);
  procedure SetZero(Template: in Fourier2; F: in out Fourier2);
  procedure SetZero(P: in ParPair; R: in RadPair; F: in out Fourier2);
  procedure Copy(P1: in Fourier2; P2: in out Fourier2);
  procedure Swap(P1,P2: in out Fourier2);

--- sets
  procedure BallAt0(R: in Radius; P: in out Fourier2);       --- 0-centered ball of radius R
  function BallAt0(R: Radius) return Fourier2;               --- 0-centered ball of radius R (random parity)
  procedure ToErr(P1,P2: in out Fourier2);                   --- P1 -> Center(P1)+P2.E
  procedure Center(P: in out Fourier2);
  procedure Center(P1: in Fourier2; P2: in out Fourier2);
  function Center(P: Fourier2) return Fourier2;
  procedure ModCenter(P: in out Fourier2);
  procedure ModCenter(P1: in Fourier2; P2: in out Fourier2);
  function ModCenter(P: Fourier2) return Fourier2;            --- zero center

--- addition and multiplication etc.
  procedure Neg(P: in out Fourier2);
  procedure Neg(P1: in Fourier2; P2: in out Fourier2);
  function "-"(P: Fourier2) return Fourier2;
  procedure Add(I: in Integer; P: in out Fourier2);
  procedure Add(S: in Scalar; P: in out Fourier2);
  procedure Add(P1: in Fourier2; P2: in out Fourier2);
  procedure Sum(P1,P2: in Fourier2; P3: in out Fourier2);
  function "+"(P1,P2: Fourier2) return Fourier2;
  procedure Sub(P1: in Fourier2; P2: in out Fourier2);
  procedure Diff(P1,P2: in Fourier2; P3: in out Fourier2);
  function "-"(P1,P2: Fourier2) return Fourier2;
  procedure Mult(R: in Flt; P,Tmp: in out Fourier2);
  procedure Mult(R: in Flt; P: in out Fourier2);
  procedure Prod(R: in Flt; P1: in Fourier2; P2: in out Fourier2);
  function "*"(R: Flt; P: Fourier2) return Fourier2;
  procedure AddProd(R: in Flt; P1: in Fourier2; P2,Tmp: in out Fourier2);
  procedure AddProd(R: in Flt; P1: in Fourier2; P2: in out Fourier2);
  procedure Mult(Q: in Rational; P: in out Fourier2);
  procedure Mult(S: in Scalar; P: in out Fourier2);
  procedure Prod(S: in Scalar; P1: in Fourier2; P2: in out Fourier2);
  function "*"(S: Scalar; P: Fourier2) return Fourier2;
  procedure AddProd(S: in Scalar; P1: in Fourier2; P2: in out Fourier2);
  procedure LMult(P1: in Fourier2; P2: in out Fourier2); -- e.g. if P1 is sparse
  procedure RMult(P1: in Fourier2; P2: in out Fourier2); -- e.g. if P2 is sparse
  procedure Mult(P1: in Fourier2; P2: in out Fourier2) renames LMult;
  procedure Prod(P1,P2: in Fourier2; P3: in out Fourier2);
  function "*"(P1,P2: Fourier2) return Fourier2;
  procedure AddProd(P1,P2: in Fourier2; P3: in out Fourier2);
  procedure Div(R: in Flt; P,Tmp: in out Fourier2);
  procedure Div(R: in Flt; P: in out Fourier2);
  procedure Quot(P1: in Fourier2; R: in Flt; P2: in out Fourier2);
  function "/"(P: Fourier2; R: Flt) return Fourier2;

--- functions
  function MaxNorm(E: FErr) return Radius;
  function MaxNorm0(C: FCoeff) return Radius;
  function MaxNorm(R: RadPair; C: FCoeff) return Radius;
  function MaxNorm(R: RadPair; F: Fourier2) return Radius;
  function MaxNorm(F: Fourier2) return Radius;
  procedure Norm(R: in RadPair; C: in FCoeff; S: in out Scalar);
  procedure Norm(R: in RadPair; F: in Fourier2; S: in out Scalar);
  function Norm(R: RadPair; F: Fourier2) return Scalar;
  function Norm(F: Fourier2) return Scalar;
  function NormOrRadius(F: Fourier2) return Scalar;
  procedure Sqr(P: in out Fourier2);
  function Sqr(P: Fourier2) return Fourier2;

--- conversion and i/o
  function Approx(P: Fourier2) return Flt;
  procedure Assign(I: in Integer; P: in out Fourier2);
  procedure Assign(Q: in Rational; P: in out Fourier2);
  procedure Assign(R: in Flt; P: in out Fourier2);
  function Scal(I: Integer) return Fourier2;
  function Scal(Q: Rational) return Fourier2;
  function Scal(R: Flt) return Fourier2;
  function Scal(S: Scalar) return Fourier2;
  procedure Enclose(R1,R2: in Flt; P: in out Fourier2);
  function Enclose(R1,R2: Flt) return Fourier2;
  procedure Show1(S: in String; T: in FType);
  procedure Show1(S: in String; F: in Fourier2);
  procedure Show2(S: in String; F1,F2: in Fourier2);
  procedure Put(F: in File_Type; H: in Fourier2; Decimal: in Boolean := False);
  procedure Get(F: in File_Type; H: in out Fourier2; Decimal: in Boolean := False);
  procedure Write(Name: in String; H: in Fourier2; Decimal: in Boolean := False);
  procedure Read(Name: in String; H: in out Fourier2; Decimal: in Boolean := False);
  
--- more
  function "*"(P1,P2: ParPair) return ParPair;
  procedure Show_Param_F2(N: in String := "");
  procedure NumRead(Name: in String; H: in out Fourier2; Decimal: in Boolean := False);
  procedure Plot(G: in Fourier2; L: Positive := 50);
  function GuessRho1(F: Fourier2) return Flt;
  function GuessRho2(F: Fourier2) return Flt;
  function GuessRho(F: Fourier2) return RadPair;
  procedure GuessRho(F: in Fourier2);

--- misc
  function Get_Precision(P: Fourier2) return Positive;
  procedure Proper_Rounding(Dummy: in Fourier2);
  procedure Free_Cache_F2;
  procedure Free_Cache(Dummy: in Fourier2);
  
--------------- other procedures
  function Power(R: Radius; I: Natural) return Radius;   --- upper bound on R**I
  procedure NormFacs(R: in Radius; W: in out Flt_Vector);
  function NumCoeffs(D1,D2: Natural; JustOdd: Boolean) return Positive;
  procedure Coeff(M,N: in Natural; F: in Fourier2; S: in out Scalar);
  function Coeff(M,N: Natural; F: Fourier2) return Scalar;
  procedure AddCoeff(M,N: in Natural; S: in Scalar; F: in out Fourier2);
  procedure AddBall(M,N: in Natural; R: in Flt; F: in out Fourier2);
  procedure AddBall(M,N: in Natural; S: in Scalar; F: in out Fourier2);
  procedure AddBall(R: in Flt; F: in out Fourier2);
  procedure AddBall(S: in Scalar; F: in out Fourier2);
  function Val(X,Y: Scalar; F: Fourier2) return Scalar;
  procedure CleanOdd1(F: in out Fourier2);
  procedure CleanOdd2(F: in out Fourier2);
  procedure CleanOdd(F: in out Fourier2);
  procedure ProjFreq1(P: in Parity; F: in out Fourier2);
  procedure ProjEvenFreq1(F: in out Fourier2);   -- even freq in 1st variable
  procedure ProjOddFreq1(F: in out Fourier2);    -- odd freq in 1st variable
  procedure ProjFreq2(P: in Parity; F: in out Fourier2);
  procedure ProjEvenFreq2(F: in out Fourier2);   -- even freq in 2nd variable
  procedure ProjOddFreq2(F: in out Fourier2);    -- odd freq in 2nd variable
  procedure Separate1(F: in out Fourier2; D: in Natural);
  procedure Separate2(F: in out Fourier2; D: in Natural);
  procedure SetRho0(F: in out Fourier2);        -- convert to F.T.R=0
  
  type Fourier2_Access is access Fourier2;
  
  FTrunc: constant Boolean;
  
private
  
  function NotExp(X: Radius) return Scalar;
  function NotExp(X: Flt) return Flt;
  procedure MaxNorm(E: in FErr; N: out Radius; Count: out Natural);
  procedure AddProd(R: in Flt; P1: in Fourier2; P2: in out Fourier2; Tmp: in out Scalar);
  procedure Signs(P1,P2: in Parity; Sum,DiffPlus,DiffMinus: out Integer);
  procedure NumProd(F1,F2: in Fourier2; F3: in out Fourier2);
  procedure MaxNorms(C: in FCoeff; E: out Flt_Matrix);
  procedure Prod(F1,F2: in Fourier2; F3: in out Fourier2; ZC1,ZE1,ZC2,ZE2: in Boolean);
  procedure Mult(P1: in Fourier2; P2,Tmp: in out Fourier2);
  function EffDeg1(C: FCoeff) return Integer;
  function EffDeg2(C: FCoeff) return Integer;
  procedure EffDegs0(C: in FCoeff; D1,D2: out Integer);
  function EffDeg1(E: FErr) return Integer;
  function EffDeg2(E: FErr) return Integer;
  procedure EffDegs0(E: in FErr; D1,D2: out Integer);
  procedure Separate1(E: in out FErr; D1,D2: in Natural);
  procedure Separate2(E: in out FErr; D1,D2: in Natural);

  DegE:     constant Positive    := IMax(DegE1,DegE2);
  SZero:    constant Scalar      := Scal(0);
  SHalf:    constant Scalar      := Scal(Half);
  SQuarter: constant Scalar      := Scal(Quarter);
  SOne:     constant Scalar      := Scal(1);
  STwo:     constant Scalar      := Scal(2);
  SThree:   constant Scalar      := Scal(3);
  SInfo:    constant Scalar_Info := Info(SZero);
  STrunc:   constant Boolean     := SInfo.IsNumeric;
  FTrunc:   constant Boolean     := SInfo.IsNumeric;
  Std:      constant Boolean     := SInfo.IsStandard;
  
  FInfo: constant Scalar_Info := (RepType    => SInfo.RepType,
                                  IsStandard => SInfo.IsStandard,
                                  IsNumeric  => SInfo.IsNumeric,
                                  IsComplex  => SInfo.IsComplex,
                                  NReps      => SInfo.NReps*(2*(DegE1+DegE2)+2),
                                  RMode      => SInfo.RMode);
  
  Task_Error: Boolean := False;
  
----- recycling
  package Alloc is new Recycling (Item => Fourier2);
  subtype LF_Pointer is Alloc.LD_Pointer;
  O: Alloc.Controlled_Protected_Data;
  Pool: Alloc.Protected_Data renames O.P;
  
end Fouriers2;
