with Globals, Ints;
use Globals, Ints;

pragma Elaborate_All (Globals,Ints);

procedure LinFlow(V0: in Vector; X: in Operator_Taylor; V: in out Vector_Taylor; DT: in Radius) is
--- solves V'(t)=X(t)V(t) with V(0)=V0
--- coefficient-errors can depend on t
--- assuming X(0..Deg) and V(0..Deg) where
  Deg: constant Positive := V'Last;
  E,Contr,R: Flt;
  S,Tmp: Vector;
begin
  Copy(V0,V(0));
  Prod(X(0),V(0),V(1));
  for K in 1 .. Deg-1 loop   -- compute V(K+1)
    SetZero(S);
    for I in 0 .. K loop
      Prod(X(I),V(K-I),Tmp);
      Add(Tmp,S);
    end loop;
    Quot(S,Flt(K+1),V(K+1));
  end loop;
  if not Numeric then
    E := Zero;
    for K in reverse Deg .. 2*Deg loop
      SetZero(S);
      for I in K-Deg .. Deg loop
        Prod(X(I),V(K-I),Tmp);
        Add(Tmp,S);
      end loop;
      E := DT*E+MaxNorm(S)/Flt(K+1);
    end loop;                        -- E is degree Deg+1 (or higher) error
    Contr := Zero;
    for K in reverse 0 .. Deg loop
      Contr := DT*Contr+OpNorm(X(K));
    end loop;
    Contr := Contr*DT/Flt(Deg+2);    -- contraction factor at degree D+1 (or higher)
    if Contr>=One then raise Sorry with "LinFlow: no contraction"; end if;
    R := E/abs(Contr-One);           -- assuming Flt is rounded up
    AddBall(DT*R,V(Deg));
  end if;
end LinFlow;
