with Ints, Globals;
use Ints;

pragma Elaborate_All (Ints,Globals);

package body Messages is

  package Txt_IO renames Ada.Text_IO;
  
  function Age(T: in Timer := Default_Timer) return Flt is
--- age in seconds
  begin
    return Flt(To_Duration(Clock-T));
  end Age;

  function Seconds(T: Timer := Default_Timer) return Integer is
  begin
    return Integer(Flt'Floor(Age(T)));
  end Seconds;

  function Minutes(T: Timer := Default_Timer) return Integer is
  begin
    return Integer(Flt'Floor(Age(T)/60.0));
  end Minutes;

  function Hours(T: Timer := Default_Timer) return Integer is
  begin
    return Integer(Flt'Floor(Age(T)/3600.0));
  end Hours;

  procedure Show_Time(P: in File_Access; T: in Timer := Default_Timer) is
    S: constant Duration := To_Duration(Clock-T);
    M: constant Integer := Integer(S)/60;
    H: constant Integer := M/60;
  begin
    if P=null then
      Int_IO.Put(H,3);
      Txt_IO.Put(":");
      Int_IO.Put(M mod 60,2);
      Txt_IO.Put(":");
      Dur_IO.Put(S-Duration(60*M),2,3);
      New_Line;
    else
      declare
        F: File_Type renames P.all;
      begin
        Int_IO.Put(F,H,3);
        Txt_IO.Put(F,":");
        Int_IO.Put(F,M mod 60,2);
        Txt_IO.Put(F,":");
        Dur_IO.Put(F,S-Duration(60*M),2,3);
        New_Line(F);
      end;
    end if;
  end Show_Time;
  
  procedure Show_Time(T: in Timer := Default_Timer) is
    use Tasks;
  begin
    Show_Time(Task_Log,T);
  end Show_Time;
  
  function Starting(Where: String) return Tracer is
    use Globals, Tasks;
    H: constant Attribute_Handle := Reference;
    P: constant File_Access := Task_Log(H);
    Level: Natural renames H.Level;
    F: File_Type renames P.all;
  begin
    if Verbosity>1 then
      for K in 1 .. Level loop Txt_IO.Put(F,". "); end loop;
      Txt_IO.Put_Line(F,"[ " & Where);
    end if;
    Level := Level+1;
    return (Controlled with T => Clock, H => H, Wait => Adjustements);
  end Starting;      
  
  procedure Adjust(T: in out Tracer) is
--- just count the number of adjustements
  begin
    T.Wait := T.Wait-1;
  end Adjust;
  
  procedure Finalize(T: in out Tracer) is
    use Globals, Tasks;
  begin
    if T.Wait /= 0 then  -- T is just being copied to/from stack
      return;            -- wait until procedure actually starts
    end if;
    declare
      Level: Natural renames T.H.Level;
    begin
      Level := Level-1;
      if Verbosity>1 then
        declare
          P: constant File_Access := Task_Log(T.H);
          F: File_Type renames P.all;
        begin
          for K in 1 .. Level loop Txt_IO.Put(F,". "); end loop;
          Txt_IO.Put(F,"] ");
          Show_Time(P,T.T);
        end;
      end if;
    end;
  end Finalize;
  
  function Seconds(T: Tracer) return Integer is
  begin
    return Integer(Flt'Floor(Age(T.T)));
  end Seconds;
  
  function Minutes(T: Tracer) return Integer is
  begin
    return Integer(Flt'Floor(Age(T.T)/60.0));
  end Minutes;

  function Hours(T: Tracer) return Integer is
  begin
    return Integer(Flt'Floor(Age(T.T)/3600.0));
  end Hours;
    
  procedure Show_Time(T: in Tracer) is
    use Tasks;
  begin
    Show_Time(Task_Log,T.T);
  end Show_Time;
  
end Messages;
