with Recycling, Flt_Vectors;

generic
  
  FreqParity: in Parity; -- even or odd x frequencies
  DimX: in Positive;     -- number of x modes

package MultiTaylors1.Phi is

  Dim2X: constant Positive := 2*DimX;
  subtype IndexX is Integer range 1 .. DimX;
  subtype Index2X is Integer range 1 .. Dim2X;

  subtype YType is Scalar_Vector(IndexX);
  subtype SMatrix is Scalar_Matrix(Index2X,Index2X);
  subtype SVector is Scalar_Vector(Index2X);
  type SVecVec is array(Index2X) of SVector; -- for Phi00

  type TMatrix is array(Power) of SMatrix;
  type TVector is array(Power) of SVector;
  type TVecVec is array(Index2X) of TVector; -- for Phi
  
  procedure Show_Param;
  procedure LinFlow(X: in TMatrix; V: in out TVecVec; DT: in Radius);
  procedure Run(DT: in Radius; M: in Scalar_Matrix; Beta2: in Scalar; C: in Scalar_Vector; Phi00: in out SMatrix);
  
  procedure Free_Cache_MT1;
  
  Flow_Parallel: Boolean := True;

private
  
  subtype Flt_Vector is Flt_Vectors.Vector;
  
  PDeg1:  constant Positive := PDeg+1;
  EDeg:   constant Positive := Choose(STrunc,PDeg1,PDeg);
  KOffset: constant Integer := -FreqParity;   -- x-frequency is Offset+2*I, I=1,2,...
  
  subtype TMatrix1 is Taylor1_Matrix(Index2X,Index2X);
  subtype TVector1 is Taylor1_Vector(Index2X);
  
  function Symplecticity(U: Scalar_Matrix) return Radius;
  function MaxNorm(V: SVector) return Radius;
  procedure AddBall(R: in Radius; V: in out SVector);
  procedure Quot(V: in SVector; R: in Flt; W: in out SVector);
  function OpNorm(A: SMatrix) return Radius;
  procedure LinFlow(X: in TMatrix; V: in out TVector; DT: in Radius; Tmp1,Tmp2: in out SVector);
  
  procedure SetZero(V: in out TVector);
  procedure T2T(A: in out TMatrix1; B: in out TMatrix);
  procedure Val(P: in TVector; X: in Scalar; Y: in out SVector);
  procedure HTilde(M: in Scalar_Matrix; Time,DT: in Radius; Y: in Scalar_Vector; H: in out TMatrix1);
  procedure MakeY(I: in Positive; Beta2,C: in Scalar; Y: in out Scalar);
  function MakeYi(Beta2: Scalar; C: Scalar_Vector) return YType;
  function MakePhi0(DT: Radius; Beta2: Scalar; C: Scalar_Vector) return TMatrix1;
  procedure ValPhi0(T,Beta2: in Scalar; C: in Scalar_Vector; V: in out SMatrix);
  
  package Alloc is new Recycling (Item => SVector);
  subtype LT_Pointer is Alloc.LD_Pointer;
  O: Alloc.Controlled_Protected_Data;
  Pool: Alloc.Protected_Data renames O.P;

end MultiTaylors1.Phi;
