with Globals, Strings, Ints, Protected_Counters;
use Globals, Strings, Ints, Protected_Counters;

package body OpNorms is
  
  function OpNormBound(Template: Fun; Parallel: Boolean := False) return Radius is
    NT: constant Natural := Reserve_Tasks(Dim-1,Parallel);
    Task_Error: Boolean := False;
    N: Radius := Zero;
    PC: Protected_Counter;
    R: array(1..Dim) of Flt;
        
    procedure Col_Loop is
      J: Integer;
      S: Scalar;
      V: PVector;
      F,G: Fun;
    begin
      for I in PRange loop SetZero(V(I)); end loop;
      loop
        PC.Next(J);
        exit when J>Dim;
        Assign(One,V(J));
        UnPack(V,Template,F);
        Op(F,G);
        Quot(Norm(G),Norm(F),S);
        R(J) := MaxNorm(S);
        if Verbosity>1 then
          Show1("OperatorNorm:" & Strng(J,4) & " ",R(J));
        end if;
        SetZero(V(J));
      end loop;
    end Col_Loop;
    
    task type Col_Task_Type is pragma Storage_Size (Stack_Size); end Col_Task_Type;
    task body Col_Task_Type is
    begin
      Col_Loop;
    exception
      when others => Task_Error := True; raise;
    end Col_Task_Type;
    
  begin
    PC.Reset(1);
    if NT=0 then
      Col_Loop;
    else
      declare
        Col_Task: array(1 .. NT) of Col_Task_Type;
      begin
        Col_Loop;
      end;
      Free_Tasks(NT);
      if Task_Error then raise Sorry with "Exception in OpNorms.OpNormBound"; end if;
    end if;
    for I in PRange loop N := RMax(R(I),N); end loop;
    return N;
  end OpNormBound;
  
  procedure Packed_Op(Template: in Fun; A: in out PMatrix; Parallel: Boolean := False) is
    NT: constant Natural := Reserve_Tasks(Dim-1,Parallel);
    Task_Error: Boolean := False;
    PC: Protected_Counter;
    
    procedure Col_Loop is
      J: Integer;
      S: Scalar;
      V: PVector;
      F,G: Fun;
    begin
      for I in PRange loop SetZero(V(I)); end loop;
      loop
        PC.Next(J);
        exit when J>Dim;
        if Verbosity>5 then Show1("Col",J); end if;
        Assign(One,V(J));
        UnPack(V,Template,F);
        S := Norm(F);
        Op(F,G);
        Pack(G,V);
        for I in PRange loop
          Quot(V(I),S,A(I,J));
          SetZero(V(I));
        end loop;
      end loop;
    end Col_Loop;
    
    task type Col_Task_Type is pragma Storage_Size (Stack_Size); end Col_Task_Type;
    task body Col_Task_Type is begin
      Col_Loop;
    exception
      when others => Task_Error := True; raise;
    end Col_Task_Type;
    
  begin
    PC.Reset(1);
    if NT=0 then
      Col_Loop;
    else
      declare
        Col_Task: array(1 .. NT) of Col_Task_Type;
      begin
        Col_Loop;
      end;
      Free_Tasks(NT);
      if Task_Error then raise Sorry with "Exception in OpNorms.Packed_Op"; end if;
    end if;
  end Packed_Op;
  
  function NormBound(Template: Fun; A: PMatrix) return Radius is
    N: Radius := Zero;
    R: Radius;
    V: PVector;
    F: Fun;
  begin
    for J in PRange loop
      for I in PRange loop
        Copy(A(I,J),V(I));
      end loop;
      UnPack(V,Template,F);
      R := MaxNorm(Norm(F));
      if N<R then N := R; end if;
    end loop;
    return N;
  end NormBound;
  
end OpNorms;
