with Rationals, Flts;
use Rationals, Flts;

pragma Elaborate_All (Rationals,Flts);

package Param is

  type Pars is
    record
      Alpha:           Rational;
      Gamma1:          Rational; --- coeff for wave
      Gamma2:          Rational; --- coeff for beam
      Kappa:           Rational;
      Deg1, Deg2:      Positive; --- highest explicit Degrees
      PDeg1, PDeg2:    Positive; --- degrees for Pack/UnPack
      EDeg1, EDeg2:    Positive; --- used in DContractNorm
      DegHat1,DegHat2: Positive; --- truncation for \hat H
      TDeg:            Positive; --- degree for Taylor in Phi2Pi
      DegHatM1:        Positive; --- truncation for \hat H matrix
      DegCheckM1:      Positive; --- truncation for \check H
      DegCheckM2:      Positive; --- truncation for \check H
      Rho1, Rho2:      Radius;   --- domain for beam solution
    end record;

  type ParsVec is array(Integer range <>) of Pars;

  Trivial: constant Pars := (0/1,0/1,0/1,0/1,1,1,1,1,1,1,1,1,1,1,1,1,Zero,Zero);

  function Modify(Template: Pars := Trivial;
                  Alpha,Gamma1,Gamma2,Kappa: Rational := 0/1;
                  Deg1,Deg2,PDeg1,PDeg2,EDeg1,EDeg2,DegHat1,DegHat2: Positive := 1;
                  TDeg,DegHatM1,DegCheckM1,DegCheckM2: Positive := 1;
                  Rho1,Rho2: Radius := Zero) return Pars;

  function HPos_Std(V: Pars) return Boolean;
  function HPos_Std(Alpha,Gamma,Kappa: Rational) return Boolean;
  function "/"(N: Integer; P: Positive) return Flt renames Frac;
  function Strng(Q: Rational; W: Positive := 2) return String;
  function Strng(P: Pars) return String;
  function Approx(Q: in Rational) return Flt;

  function Fix_Name(P: Pars) return String;
  function Full_Name(P: Pars) return String;

  Num_Bits: Positive := 512;

end Param;
