with Ada.Text_IO, Messages, Strings, Ints, Flts, Globals, Reps, Rationals, Flt_Vectors, Param;
use Ada.Text_IO, Messages, Strings, Ints, Flts, Globals, Reps, Rationals, Flt_Vectors, Param;
with Matrices, Vectors, ScalVectors, Fouriers2, Taylors1, MultiTaylors1,MultiTaylors1.Phi;

pragma Elaborate_All (Ada.Text_IO,Messages,Strings,Ints,Flts,Globals,Reps);
pragma Elaborate_All (Matrices,Vectors,ScalVectors,Fouriers2,Taylors1,MultiTaylors1,MultiTaylors1.Phi);

with Reps.Balls;
use Reps.Balls;
pragma Elaborate_All (Reps.Balls);

procedure Phi2Pi(V: in Pars; FreqParity: in Parity; EigenBounds: in Boolean) is

  subtype Scalar is Ball;

  NumDigits: constant Positive := Set_Precision(768);
  PDeg: constant Positive := V.TDeg;
  DimX: constant Positive := (V.DegHat2+FreqParity)/2;

  package SM is new Matrices (Scalar => Scalar);
  package SV is new Vectors (Component => Scalar);
  package SSV is new ScalVectors (Scalar => Scalar, SM => SM, SV => SV);
  package SFouriers2 is new Fouriers2 (Deg1 => V.Deg1, Deg2 => V.Deg2, Scalar => Scalar, SM => SM, SV => SV);
  package STaylors1 is new Taylors1 (PDeg => PDeg, Scalar => Scalar, SM => SM, SV => SV);
  package MTaylors1 is new MultiTaylors1 (PDeg => PDeg, Scalar => Scalar, SM => SM, SV => SV,
                                          SSV => SSV, ST1 => STaylors1);

  package MTaylors1_Phi is new MTaylors1.Phi (DimX => DimX, FreqParity => FreqParity);

  use SM, SFouriers2, MTaylors1, MTaylors1_Phi;

  Postfix3:  constant String := Fix_Name(V);
  Postfix3P: constant String := Postfix3 & ".P" & Strng(FreqParity);
  Postfix4:  constant String := Full_Name(V);
  Postfix4P: constant String := Postfix4 & ".P" & Strng(FreqParity);
  DT:        constant Radius := Half**5;
  Beta2:     constant Scalar := Scal(Sqr(V.Gamma1/V.Alpha));

  C0: Scalar;
  D: Flt_Vector(IndexX);
  C: YType;
  H: Fourier2;
  Phi00: SMatrix;

begin
  Show_Param_F2;
  Show_Param;
  Show1("alpha =",V.Alpha);
  Show1("gamma =",V.Gamma1);
  Show1("kappa =",V.Kappa);

  Read(Input_Dir & "fixh_" & Postfix3,H);
  Read(Input_Dir & "c_" & Postfix3,C0);
  Read(Input_Dir & "d_" & Postfix3P,D);
  
  
  for M in C'Range loop C(M) := C0-Scal(D(M)); end loop;
  Run(DT,H.C,Beta2,C,Phi00);
  Write(Output_Dir & "phi.D-." & Postfix4P,Phi00);
  
  if EigenBounds then
     for M in C'Range loop C(M) := C0+Scal(D(M)); end loop;
     Run(DT,H.C,Beta2,C,Phi00);
     Write(Output_Dir & "phi.D+." & Postfix4P,Phi00);
  end if;
  
end Phi2Pi;
