with Scalars, Flts, Bisection;
use Scalars, Flts;

pragma Elaborate_All (Scalars,Flts,Bisection);

generic
  
  type Scalar is private;
  
  with function Info(Dummy: Scalar) return Scalar_Info is <>;
  with function IsZero(S: Scalar) return Boolean is  <>;
  with procedure SetZero(S: in out Scalar) is <>;
  with procedure Swap(S1,S2: in out Scalar) is <>;
  with procedure Copy(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Center(S: in out Scalar) is <>;
  with procedure Neg(S: in out Scalar) is <>;
  with procedure Neg(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Add(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Sub(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Sum(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure Diff(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure Mult(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Prod(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure AddProd(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure Inv(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Quot(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure Roots2(B,C: in Scalar; U1,U2,V: in out Scalar) is <>;
  with procedure Show1(N: in String; S: in Scalar; Whatever: in Boolean := True) is <>;
  with procedure Show2(N: in String; S1,S2: in Scalar; Whatever: in Boolean := True) is <>;
  
  with function Scal(R: Flt) return Scalar is <>;
  with function Sign(S: Scalar) return Integer is <>;
  with procedure Mult(R: in Flt; S: in out Scalar) is <>;
  with procedure Assign(R: in Flt; S: in out Scalar) is <>;
  with function Approx(S: Scalar) return Flt is <>;
  with function MaxNorm(S: Scalar) return Radius is <>;
  with procedure BallAt0(R: in Flt; S: in out Scalar) is <>;
  
  type Vector is array(Integer range <>) of Scalar;
  type Matrix is array(Integer range <>, Integer range <>) of Scalar;
  
package PolyRoots is
  
  subtype Vec2 is Vector (1..2);
  subtype Mat2 is Matrix (1..2,1..2);
  subtype Poly is Vector;
  subtype Cheby is Vector;
  
  procedure PolyVal(P: in Poly; S: in Scalar; T: in out Scalar);         -- T := P(S)
  procedure PolyVal(P: in Poly; Sr,Si: in Scalar; Tr,Ti: in out Scalar); -- T := P(S)  for real polynomials
  procedure PolyValDer(P: in Poly; S: in Scalar; T: in out Scalar);      -- T := P'(S)
  procedure LMultiply(W: in Scalar; P: in out Poly);                     -- P(x) := (x-W)*P(x)
  procedure LMultiply(A,B: in Scalar; P: in out Poly);                   -- P(x) := (A*x+B)*P(x)
  procedure QMultiply(U,V,C,D: in Scalar; P: in out Poly);               -- P(x) := (x*x+U*x+V)*P(x)+(C*x+D)
  procedure QProduct(U,V,C,D: in Scalar; Q: in Poly; P: in out Poly);    -- P(x) := (x*x+U*x+V)*Q(x)+(C*x+D)
  procedure QMultiply(U,V: in Vector; A,B: in Scalar; P: in out Poly);   -- P(x) := (A*x+B)*Prod_k(x*x+U(K)*x+V(K))
  
  procedure FindRealRoot(P: in Poly; W: in out Scalar; Err: out Radius);
  procedure FindRealRoots(P: in Poly; W: in out Vector);
  procedure Dec_MaxNorm_Sort(V: in out Vector);
  procedure Sort(V: in out Vector) renames Dec_MaxNorm_Sort;
  procedure LDivision(P: in Poly; W: in Scalar; Q: in out Poly; R: in out Scalar);      -- divivion of P(x) by by (x-W)
  procedure LDivision(P: in out Poly; W: in Scalar; R: in out Scalar);                  -- divivion of P(x) by by (x-W)
  procedure QDivision(P: in Poly; U,V: in Scalar; Q: in out Poly; C,D: in out Scalar);  -- division of P(x) by (x*x+U*x+V)
  procedure QDivision(P: in Poly; U,V: in Scalar; Q: in out Poly);                      -- division of P(x) by (x*x+U*x+V)
  procedure FindQFactor(P: in Poly; U,V: in out Scalar; Q: in out Poly; Err: out Radius);  -- needs approx U,V
  procedure FindQFactors(P: in Poly; U,V: in out Vector; A,B: in out Scalar);              -- needs approx U,V
  procedure Roots(Deg: in Positive; U,V: in Vector; A,B: in Scalar; Wr,Wi: in out Vector); -- needs approx U,V
  
--- symplectic polynomials: roots come in pairs (r,1/r)
  procedure Clenshaw(C: in Cheby; P: in out Poly);         -- convert finite Cheby series to polynomial
  procedure SymPolyCheby(P: in Poly; C: in out Cheby);     -- Cheby assiciated with P
  procedure SymCosPoly(P: in Poly; Q: in out Poly);        -- roots of Q are (r+1/r)/2
  function Right_Sign(P: Poly) return Integer;             -- sign of P(0)*P(1)
  function Left_Sign(P: Poly) return Integer;              -- sign of P(0)*P(-1)
  procedure CircleRoots(P: in Poly; W: in out Vector);     -- needs approximate W (real parts of roots)
--- Bairstow
  procedure Sym_FindQFactor(P: in Poly; U0: in out Scalar; Q: in out Poly; Err: out Radius); -- needs approx U
  procedure Sym_FindQFactors(P: in Poly; U: in out Vector);                                  -- needs approx U
  procedure Sym_Roots(Deg: in Positive; U: in Vector; Wr,Wi: in out Vector);                 -- needs approx U
  
--- purely numerical (note: eps is only for bisection)
  procedure NumRealRoot(P: in Poly; WMin,WMax: in Flt; W: in out Scalar; Eps: in Radius);  -- improves result of RoughRealRoot
  procedure NumRealRoots(P: in Poly; WMin,WMax: in Flt; W: in out Vector; Eps: in Radius); -- improves result of RoughRealRoots
  procedure RoughRealRoot(P: in Poly; WMin,WMax: in Flt; W: in out Scalar; Err: in out Radius);              -- bisection
  procedure RoughRealRoots(P: in Poly; WMin,WMax: in Flt; W: in out Vector; Q: in out Poly; Eps: in Radius); -- bisections
  procedure NumCircleRoots(P: in Poly; W: in out Vector; Eps: in Radius);                  -- first SymCosPoly then NumRealRoots
--- too cheap
  procedure RoughQFactor(P: in Poly; U,V: in out Scalar; Q: in out Poly; Eps: in Radius);
  procedure RoughQFactors(P: in Poly; U,V: in out Vector; A,B: in out Scalar);
  procedure RoughRoots(P: in Poly; Wr,Wi: in out Vector);
  
--- for testing etc
  procedure Quot(B: in Vec2; A: in Mat2; X: in out Vec2);       -- solve A*X=B
  procedure Quot(B,A: in Mat2; X: in out Mat2);                 -- solve A*X=B
  procedure Product(P,Q: in Poly; R: in out Poly);              -- R := P*Q
  procedure SymProd(P1: in Poly; P2: in out Poly);              -- P2(x) := (x^D)P1(1/x)*P1(x)
  procedure Chebyshev(N: in Natural; P: in out Poly);           -- N-th Chebyshev polynomial
  function SymErr(P: Poly) return Radius;                       -- measure of asymmetry
  
  procedure RandomReal(W: in out Vector; P: in out Poly; Fac: in Flt := One);
  procedure RandomPoly(U,V: in out Vector; P: in out Poly; Fac: in Flt := Three);
  procedure RandomSymPoly(U: in out Vector; P: in out Poly; Fac: in Flt := Two);
  
--- some Newton stuff
  generic
    with procedure TF(S: in Scalar; T,A: in out Scalar);
  procedure Find01(S: in out Scalar; Steps: in Natural := 64; Verb: in Natural := 0);
  
  generic
    with procedure F(S: in Scalar; T: in out Scalar);
    with procedure DF(S: in Scalar; A: in out Scalar);
  procedure FindZero1(S: in out Scalar; Steps: in Natural := 64; Verb: in Natural := 0);
  
  generic
    with procedure TF(V: in Vec2; W: in out Vec2; A: in out Mat2);
  procedure Find02(V: in out Vec2; Steps: in Natural := 64; Verb: in Natural := 0);
  
  generic
    with procedure F(V: in Vec2; W: in out Vec2);
    with procedure DF(V: in Vec2; A: in out Mat2);
  procedure FindZero2(V: in out Vec2; Steps: in Natural := 64; Verb: in Natural := 0);
  
  package Bisect is new Bisection(Scalar => Scalar);
  
  NSign: Positive renames Bisect.NSign;
  NFind: Positive renames Bisect.NFind;
  
  Off_Circle: exception; -- used by CircleRoots
  
private
  
  function EffDeg(P: Poly) return Integer;
  procedure Copy(P: in Poly; Q: in out Poly);
  procedure PolyVal(P: in Poly; Deg: in Natural; S: in Scalar; T: in out Scalar);
  procedure PolyValDer(P: in Poly; Deg: in Natural; S: in Scalar; T: in out Scalar);
  
  procedure Add(I: in Integer; S: in out Scalar);
  procedure Center(V: in out Vec2);
  procedure Add(U: in Vec2; V: in out Vec2);
  procedure Sub(U: in Vec2; V: in out Vec2);
  procedure Neg(A: in out Mat2);
  procedure Center(A: in out Mat2);
  procedure Copy(A: in Mat2; B: in out Mat2);
  procedure Add(A: in Mat2; B: in out Mat2);
  procedure Add(I: in Integer; A: in out Mat2);
  procedure Prod(A,B: in Mat2; C: in out Mat2);
  procedure InvDet(A: in Mat2; S,Tmp: in out Scalar);
  function MaxNorm(V: Vec2) return Radius;
  function MaxNorm(A: Mat2) return Radius;
  procedure BallAt0(R: in Flt; V: in out Vec2);
  procedure BallAt0(R: in Flt; A: in out Mat2);
  
  function GT_MaxNorm(A,B: Scalar) return Boolean;
  procedure Root(P0,P1: in Scalar; W: in out Scalar);
  procedure RealRoots2(P0,P1,P2: in Scalar; W1,W2: in out Scalar);
  procedure Bairstow0(P: in Poly; Deg: in Integer; S: in Vec2; Q: in out Poly; FS: in out Vec2);  -- QDivide map F
  procedure Bairstow1(Q: in Poly; Deg: in Integer; S: in Vec2; R: in out Poly; DFS: in out Mat2); -- derivative DF
  procedure FindQFactor2(P0,P1,P2: in Scalar; U,V: in out Scalar; Q: in out Poly);
  procedure Sym_FindQFactor2(P0,P1,P2: in Scalar; U: in out Scalar; Q: in out Poly);
    
  Scal_Zero: constant Scalar      := Scal(Zero);
  Scal_Half: constant Scalar      := Scal(Half);
  Scal_One:  constant Scalar      := Scal(One);
  Scal_Two:  constant Scalar      := Scal(Two);
  SInfo:     constant Scalar_Info := Info(Scal_Zero);
  STrunc:    constant Boolean     := SInfo.IsNumeric;
  
end PolyRoots;
