with Globals, Logicals, Rationals, Param, Ints, Flts, Strings;
use Globals, Logicals, Rationals, Param, Ints, Flts, Strings;
with Check_Fix, Save_Data, Phi2Pi, Eigen_Circle, Barrier;

pragma Elaborate_All (Globals,Logicals,Rationals,Param,Ints,Flts);
pragma Elaborate_All (Check_Fix,Save_Data,Phi2Pi,Eigen_Circle,Barrier);

procedure Run_All is
   
      V: constant Pars := Modify( Alpha => 5/4,
                               Gamma1 => 7/8,      ----- coefficient for wave
                               Gamma2 => 1/1,      ----- coefficient for beam
                               Kappa => 1/2,
                               Deg1 => 141,
                               Deg2 => 95,
                               PDeg1 => 15,
                               PDeg2 => 11,
                               EDeg1 => 25,
                               EDeg2 => 7,
   			       DegHat1 => 41,
                               DegHat2 => 19,
   			       DegHatM1 => 35,
   			       DegCheckM1 => 35,
   			       DegCheckM2 => 11,
                               TDeg => 80,
                               Rho1 => 1/8,
                               Rho2 => 1/8);

   subtype Solutions is Integer range 1..2;

   PV:   constant ParsVec(Solutions) := (V,                                     --    stable
					 Modify(V, Alpha => 14/11, DegHat1 => 111, DegHatM1 => 80,
						DegCheckM1 => 80, TDeg => 120)  -- unstable
					 );
					 
   NT:   constant Positive := GetArg;       --- number of concurrent tasks
   J :   constant Positive := GetArg;       --- number of solution
   EigenBounds: constant Boolean := False;  --- set to True to get tight bounds on the eigenvalues
   
begin

   Verbosity := 1;
   NTasks := NT;
   Check_Fix(PV(J));
   Show0("Check_Fix done");
   for I in Parity loop
      Show1("Starting parity ",I);
      Save_Data(PV(J),I);
      Phi2pi(PV(J),I,EigenBounds);
      Show0("Phi2pi done");
      Eigen_Circle(PV(J),I,EigenBounds);
      Barrier(PV(J),I);
      Show0("Barrier done");
   end loop;
   
end Run_All;
