with Ada.Text_IO, Strings, Ints, Flts, Globals, Param, Reps, Rationals, Flt_Vectors, Param;
use Ada.Text_IO, Strings, Ints, Flts, Globals, Param, Reps, Rationals, Param;
with Matrices, Vectors, ScalVectors, Fouriers2, Fouriers2.FlokD;
with Flok_Intervals;
use Flok_Intervals;

pragma Elaborate_All (Ada.Text_IO,Strings,Ints,Globals,Param,Reps,Rationals,Flt_Vectors);
pragma Elaborate_All (Matrices,Vectors,ScalVectors,Fouriers2.FlokD);

with Reps.Balls;
use Reps.Balls;
pragma Elaborate_All (Reps.Balls);

procedure Save_Data(V: in Pars; FreqParity: in Parity) is

  subtype Scalar is Ball;

  package SM is new Matrices (Scalar => Scalar);
  package SV is new Vectors (Component => Scalar);
  package SSV is new ScalVectors (Scalar => Scalar, SM => SM, SV => SV);
  package SFouriers2 is new Fouriers2 (Deg1 => V.Deg1, Deg2 => V.Deg2, Scalar => Scalar, SM => SM, SV => SV);
  package SFouriers2_FlokD is new SFouriers2.FlokD (FreqParity => FreqParity, DegHat1=>V.DegHat1, DegHat2=>V.DegHat2);

  use SFouriers2,SFouriers2_FlokD,SM,SV,SSV;

  Postfix3:  constant String := Fix_Name(V);
  Postfix3P: constant String := Postfix3 & ".P" & Strng(FreqParity);
  Postfix4:  constant String := Full_Name(V);
  Postfix4P: constant String := Postfix4 & ".P" & Strng(FreqParity);

  procedure Make_H(H: in out Fourier2) is
    AA3: constant Scalar := Scal(3*Sqr(1/V.Alpha));
    F: Fourier2;
  begin
    Read(Input_Dir & "fix_" & Postfix3,F);
    ProjOddFreq1(F);
    ProjOddFreq2(F);
    Prod(F,F,H);
    SetRho0(H);
    ProjEvenFreq1(H);
    ProjEvenFreq2(H);
    Mult(AA3,H);
  end Make_H;

  procedure Split_Off_C(H: in out Fourier2; C: in out Scalar) is
    C0: Scalar renames H.C(0,0);
  begin
     C := Center(C0)+Scal(V.Kappa/Sqr(V.Alpha));
     Write(Output_Dir & "c_" & Postfix3,C);
     Show1("C",MaxNorm(C));
     ModCenter(C0);
  end Split_Off_C;

  procedure Split_Off_D(H: in out Fourier2; DBound: out Radius) is
    use Flt_Vectors;
    L: constant Integer := (V.Deg1+1)/2;
    D: Flt_Vector(1..L);
  begin
     MakeD(H,D);                         -- also removes errors from H
     Write(Output_Dir & "d_" & Postfix3P,D);
     Write(Output_Dir & "fixh_" & Postfix3,H);
     Show1("|H|",MaxNorm(H));
     DBound := D(L);
  end Split_Off_D;

  procedure Make_High_Eta(H: in Fourier2; C: in Scalar; DBound: in Radius) is
    Beta1: constant Rational := V.Gamma1/V.Alpha;
    E:  constant F_Intervals := High_Eta(Beta1,C,DBound);
  begin
    Write(Output_Dir & "hi-eta_" & Postfix4P,E);
  end Make_High_Eta;

  DBound: Radius;
  C: Scalar;
  H: Fourier2;

begin
  Set_Precision(768);
  FProd_Parallel := True;
  Show_Degs;

  Make_H(H);
  Split_Off_C(H,C);
  Split_Off_D(H,DBound);
  if DBound>Inf(C) then raise Sorry with "DBound>C"; end if;
  Make_High_Eta(H,C,DBound);
  
end Save_Data;
