with Strings, Ints, Flts.Plotting;
use Strings, Ints;

pragma Elaborate_All (Strings,Flts.Plotting);

package body ScalVectors.Phi is

  procedure IndexG(Vr,Vi: in Vector; S: in out Scalar) is
    L: constant Positive := Vr'Last/2;
    I2: Integer;
    N,Tmp: Scalar;
  begin
    SetZero(S);
    if IsZero(Vi) then return; end if;
    SetZero(N);
    for I in 1 .. 2*L loop
      Prod(Vr(I),Vr(I),Tmp);
      Add(Tmp,N);
      Prod(Vi(I),Vi(I),Tmp);
      Add(Tmp,N);
    end loop;
    for I1 in 1 .. L loop
      I2 := I1+L;
      Prod(Vr(I1),Vi(I2),Tmp);
      Add(Tmp,S);
      Prod(Vr(I2),Vi(I1),Tmp);
      Sub(Tmp,S);
    end loop;
    Prod(-Two,S,Tmp);
    Quot(Tmp,N,S);
  end IndexG;

  function ZeroColumn(J: Integer; A: Matrix) return Boolean is
  begin
    for I in A'Range(1) loop
      if not IsZero(A(I,J)) then return False; end if;
    end loop;
    return True;
  end ZeroColumn;

  procedure IndexG(J: in Positive; Vr,Vi: in Matrix; S,Tmp1,Tmp2: in out Scalar) is
    L: constant Positive := Vr'Last(1)/2;
    I2: Integer;
    N: Scalar renames Tmp1;
  begin
    SetZero(S);
    if ZeroColumn(J,Vi) then return; end if;
    SetZero(N);
    for I in 1 .. 2*L loop
      Prod(Vr(I,J),Vr(I,J),Tmp2);
      Add(Tmp2,N);
      Prod(Vi(I,J),Vi(I,J),Tmp2);
      Add(Tmp2,N);
    end loop;
    for I1 in 1 .. L loop
      I2 := I1+L;
      Prod(Vr(I1,J),Vi(I2,J),Tmp2);
      Add(Tmp2,S);
      Prod(Vr(I2,J),Vi(I1,J),Tmp2);
      Sub(Tmp2,S);
    end loop;
    Prod(-Two,S,Tmp2);
    Quot(Tmp2,N,S);
  end IndexG;

  procedure IndexG(J: in Positive; Vr,Vi: in Matrix; S: in out Scalar) is
    Tmp1,Tmp2: Scalar;
  begin
    IndexG(J,Vr,Vi,S,Tmp1,Tmp2);
  end IndexG;

  procedure PosIndexG(Wr,Wi: in out Vector; Vr,Vi: in out Matrix) is
    --- change signs of Wi(J) and Vi(*,J), if necessary, to get nonnegative index
    S,Tmp1,Tmp2: Scalar;
  begin
    for J in Vr'Range(2) loop
      IndexG(J,Vr,Vi,S,Tmp1,Tmp2);
      if Sign(S)<0 then
        Neg(Wi(J));
        for I in Vr'Range(1) loop Neg(Vi(I,J)); end loop;
      end if;
      if Verbosity>1 then
        Show2(Strng(J,3) & " Wr,Wi",Approx(Wr(J)),Approx(Wi(J)),False);
        Show1("  Index",Approx(S));
      end if;
    end loop;
  end PosIndexG;

  -----------------------------

  procedure Make0(Eta: in Scalar; A: out F_Interval) is
  begin
    A.Num_Lower := Inf(Eta);
    A.Num_Upper := Sup(Eta);
    A.Den := 1;
    A.Finite := True;
  end Make0;

  function Low_Eta(Wr,Wi: Vector; Vr,Vi: Matrix) return F_Intervals is
    InvTwoPi: constant Scalar := Half_Scalar/Pi;
    L: constant Positive := Wr'Last;
    S,Tmp1,Tmp2: Scalar;
    F: F_Intervals(1 .. 2*L);
  begin
    for J in 1 .. L loop
      if MaxNorm(Wr(J))<One then
        ArcCos(Wr(J),S);
        Prod(InvTwoPi,S,Tmp1);
        Diff(One_Scalar,Tmp1,Tmp2);
        if Sign(Wi(J))<0 then
          Swap(Tmp1,Tmp2);
        end if;
        declare
          F1: F_Interval renames F(2*J-1);
          F2: F_Interval renames F(2*J);
        begin
          Make0(Tmp1,F1);
          Make0(Tmp2,F2);
          IndexG(J,Vr,Vi,S,Tmp1,Tmp2);
          F1.Index := Sign(S);
          F2.Index := -F1.Index;
        end;
      end if;
    end loop;
    Check_Intervals(F);
    if Verbosity>4 then Show1("low_eta",F); end if;
    return F;
  end Low_Eta;

  function Separating_Eta(Low,High: F_Intervals) return FltVec is
    Eps: constant Radius := 1.0E-6;
    Z: Parity;
    L: Integer;
    R: Flt;
    V: F_Intervals(1 .. Low'Length+High'Length) := Low & High;
  begin
    Gaps(V,L,Z);
    declare
      G: FltVec(1 .. L+Z);
    begin
      if Z=1 then G(1) := Zero; end if;
      for I in 1 .. L loop
        R := Half*(V(I).Num_Lower+V(I).Num_Upper);
        if abs(R-Half)<Eps then
          G(I+Z) := Half;
        else
          G(I+Z) := R;
        end if;
      end loop;
      return G;
    end;
  end Separating_Eta;

  ----------- plotting

  procedure Plot_Circle(Name: in String; Steps: in Positive := 256) is
    use Flts.Plotting, Flt_EF;
    TwoPi: constant Flt := 6.283185307179586476925286766559;
    DT: constant Flt := TwoPi/Flt(Steps);
    T: Flt;
    F: File_Type;
  begin
    Create(F,Out_File,Name);
    for N in 0 .. Steps loop
      T := Flt(N)*DT;
      Plot2(F,Cos(T),Sin(T));
    end loop;
    Close(F);
  end Plot_Circle;

  procedure Plot_Eigen_Circle(Stem: in String; Wr: in Vector) is
    use Flts.Plotting, Flt_EF;
    Name: constant String := Stem & ".wr.xy";
    X,Y: Flt;
    F: File_Type;
  begin
    Show0("Writing " & Name);
    Create(F,Out_File,Name);
    for J in Wr'Range loop
      X := Approx(Wr(J));
      if abs(X)>One then
        Plot2(F,X,Zero);
        Plot2(F,One/X,Zero);
      else
        Y := Sqrt(One-X*X);
        Plot2(F,X, Y);
        Plot2(F,X,-Y);
      end if;
    end loop;
    Close(F);
  end Plot_Eigen_Circle;

  procedure Plot_Eigen_Pos(Stem: in String; Wr,Wi: in Vector) is
    use Flts.Plotting;
    Name: constant String := Stem & ".wpos.xy";
    X,Y: Flt;
    F: File_Type;
  begin
    Show0("Writing " & Name);
    Create(F,Out_File,Name);
    for J in Wr'Range loop
      X := Approx(Wr(J));
      if abs(X) <= One then
        Y := Approx(Wi(J));
        Plot2(F,X,Y);
      end if;
    end loop;
    Close(F);
  end Plot_Eigen_Pos;

  procedure Plot_Eigen_Neg(Stem: in String; Wr,Wi: in Vector) is
    use Flts.Plotting;
    Name: constant String := Stem & ".wneg.xy";
    X,Y: Flt;
    F: File_Type;
  begin
    Show0("Writing " & Name);
    Create(F,Out_File,Name);
    for J in Wr'Range loop
      X := Approx(Wr(J));
      if abs(X) <= One then
        Y := Approx(Wi(J));
        Plot2(F,X,-Y);
      end if;
    end loop;
    Close(F);
  end Plot_Eigen_Neg;

  function Count_Eigen_Zero(Wr: Vector) return Natural is
    Count: Natural := 0;
  begin
    for J in Wr'Range loop
      if abs(Approx(Wr(J)))>One then Count := Count+1; end if;
    end loop;
    return Count;
  end Count_Eigen_Zero;

  procedure Plot_Eigen_Zero(Stem: in String; Wr,Wi: in Vector) is
  begin
    if Count_Eigen_Zero(Wr)>0 then
      declare
        use Flts.Plotting;
        Name: constant String := Stem & ".wzero.xy";
        X: Flt;
        F: File_Type;
      begin
        Create(F,Out_File,Name);
        for J in Wr'Range loop
          X := Approx(Wr(J));
          if abs(X)>One then
            Plot2(F,X,Zero);
            Plot2(F,One/X,Zero);
          end if;
        end loop;
        Close(F);
      end;
    end if;
  end Plot_Eigen_Zero;

end ScalVectors.Phi;
