with Ada.Text_IO, Scalars, Globals, Logicals, Flts, Flt_Vectors, Matrices, Vectors, Vectors.Ops, PolyRoots;
use Ada.Text_IO, Scalars, Globals, Logicals, Flts;

pragma Elaborate_All (Ada.Text_IO,Scalars,Globals,Logicals,Flts,Flt_Vectors,Matrices,Vectors,Vectors.Ops,PolyRoots);

generic

  type Scalar is private;

  with function Info(Dummy: Scalar) return Scalar_Info is <>;
  with function "="(S1,S2: Scalar) return Boolean is <>;
  with function IsZero(S: Scalar) return Boolean is  <>;
  with function IsReal(S: Scalar) return Boolean is  <>;
  with procedure SetZero(S: in out Scalar) is <>;
  with procedure Assign(K: in Integer; S: in out Scalar) is <>;
  with procedure Assign(R: in Flt; S: in out Scalar) is <>;
  with procedure Copy(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Swap(S1,S2: in out Scalar) is <>;
  with procedure Center(S: in out Scalar) is <>;
  with procedure BallAt0(R: in Flt; S: in out Scalar) is <>;
  with function Inf(S: Scalar) return Flt is <>;
  with function Sup(S: Scalar) return Flt is <>;
  with function "abs"(S: Scalar) return Scalar is <>;
  with procedure Neg(S: in out Scalar) is <>;
  with procedure Neg(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Add(I: in Integer; S: in out Scalar) is <>;
  with procedure Add(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Sum(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure Sub(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Diff(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure Mult(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Prod(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure Quot(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure AddProd(S1,S2: in Scalar; S3,Tmp: in out Scalar) is <>;
  with procedure AddProd(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure Real_Part(S: in out Scalar) is <>;
  with procedure Imag_Part(S: in out Scalar) is <>;
  with procedure Adjoint(S: in out Scalar) is <>;
  with procedure Adjoint(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Show1(N: in String; S: in Scalar; Whatever: in Boolean := True) is <>;
  with procedure Show2(N: in String; S1,S2: in Scalar; Whatever: in Boolean := True) is <>;
  with procedure Put(F: in File_Type; S: in Scalar; Decimal: in Boolean := False) is <>;
  with procedure Get(F: in File_Type; S: in out Scalar; Decimal: in Boolean := False) is <>;
  with procedure Proper_Rounding(Dummy: in Scalar) is <>;

  with package SM is new Matrices (Scalar => Scalar);
  with package SV is new Vectors (Component => Scalar);

  with function "-"(S: Scalar) return Scalar is <>;
  with function "+"(S1,S2: Scalar) return Scalar is <>;
  with function "-"(S1,S2: Scalar) return Scalar is <>;
  with procedure Prod(R: in Flt; S1: in Scalar; S2: in out Scalar) is <>;
  with function "*"(R: Flt; S: Scalar) return Scalar is <>;
  with function "*"(S1,S2: Scalar) return Scalar is <>;
  with procedure Inv(S1: in Scalar; S2: in out Scalar) is <>;
  with function Inv(S: Scalar) return Scalar is <>;
  with function "**"(S: Scalar; I: Integer) return Scalar is <>;
  with function Exp(S: Scalar) return Scalar is <>;
  with function Log(S: Scalar) return Scalar is <>;
  with procedure Max(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Norm(S1: in Scalar; S2: in out Scalar) is <>;
  with function MaxNorm(S: Scalar) return Radius is <>;
  with function Sqr(S: Scalar) return Scalar is <>;
  with procedure Sqrt(S1: in Scalar; S2: in out Scalar) is <>;
  with function Sqrt(S: Scalar) return Scalar is <>;
  with procedure Roots2(B,C: in Scalar; U1,U2,V: in out Scalar) is <>;
  with procedure Roots3(B,C,D: in Scalar; U0,U1,U2,V: in out Scalar) is <>;
  with function Scal(R: Flt) return Scalar is <>;
  with function Approx(S: Scalar) return Flt is <>;
  with function Sign(S: Scalar) return Integer is <>;
  with procedure Mult(R: in Flt; S: in out Scalar) is <>;

package ScalVectors is

  package SV_Ops is new SV.Ops (Scalar => Scalar, SM => SM);
  package SPR is new PolyRoots (Scalar => Scalar, Vector => SV.Vector, Matrix => SM.Matrix);
  use SM,SV;

  procedure Norm1(V: in Vector; S: in out Scalar);                --- sum
  function Norm1(V: Vector) return Scalar;                        --- sum
  procedure WNorm1(W,V: in Vector; S: in out Scalar);             --- weighted sum
  procedure WNorm1(W: in RadVec; V: in Vector; S: in out Scalar); --- weighted sum
  function WNorm1(W: RadVec; V: Vector) return Scalar;            --- weighted sum
  procedure WBallAt0(W: in RadVec; R: in Flt; V: in out Vector);
  procedure Norm1(A: in Matrix; S: in out Scalar);                --- sum
  function Norm1(A: Matrix) return Scalar;                        --- sum
  function MaxNorm2(A: Matrix; Pow4: Positive := 2) return Radius;
  procedure WNorm1(W: in Vector; A: in Matrix; S: in out Scalar); --- weighted sum
  procedure WNorm1(W: in RadVec; A: in Matrix; S: in out Scalar); --- weighted sum
  function WNorm1(W: RadVec; A: Matrix) return Scalar;            --- weighted sum
  procedure WBallAt0(W: in RadVec; R: in Flt; A: in out Matrix);
  function Norm2(V: Vector) return Scalar;
  function NormI(V: Vector) return Scalar;                        --- max

  subtype FltVec is Flt_Vectors.Vector;
  type Diag is new Vector;

  procedure Center(V: in out Vector);
  procedure Center(A: in out Matrix);
  procedure Dot(V1,V2: in Vector; S: in out Scalar);
  function Dot(V1,V2: Vector) return Scalar;         --- linear in both args
  function "*"(V1,V2: Vector) return Scalar;         --- antilinear in V1
  function "*"(S: Scalar; U: FltVec) return Vector;
  procedure Truncate(Eps: in Radius; V: in out Vector);
  procedure Truncate(Eps: in Radius; A: in out Matrix);
  procedure SetRow(I: in Integer; V: in Vector; A: in out Matrix);
  procedure SetColumn(J: in Integer; V: in Vector; A: in out Matrix);
  procedure GetRow(I: in Integer; A: in Matrix; V: in out Vector);
  function GetRow(I: Integer; A: Matrix) return Vector;
  procedure GetColumn(J: in Integer; A: in Matrix; V: in out Vector);
  function GetColumn(J: Integer; A: Matrix) return Vector;
  procedure Symmetrize(A: in out Matrix);
  function AntiSymm(A: in Matrix) return Matrix;
  procedure Orthogonalize(A: in out Matrix);                           -- Gram-Schmidt
  procedure LDLdecomp(A: in Matrix; L: in out Matrix; D: in out Diag); -- LDL decomp of Hermitian A
  function IsPositive(A: Matrix) return Logical;                       -- for symmetric A
  function Determinant(A: Matrix) return Scalar;
  procedure Solve(B: in Vector; A: in Matrix; X: in out Vector);       -- Solve A*X=B
  procedure Invert(A: in out Matrix);                                  -- Invert A

  subtype EigenPolys is Matrix; -- range (0 .. N,0 .. N)

  procedure QR(A,Q: in out Matrix);                                       -- A -> QA by Householder reflections
  procedure Francis_Upper(A,Q: in out Matrix; Symm: in Boolean := False); -- A -> Q^tAQ by Householder reflections
  procedure Francis_Lower(A,Q: in out Matrix; Symm: in Boolean := False); -- A -> Q^tAQ by Householder reflections
  procedure Francis(A,Q: in out Matrix; Symm: in Boolean := False) renames Francis_Lower;
  procedure DeterminantTri(A: in Matrix; D: in out Scalar);         -- determinant of a tridiagonal matrix
  function DeterminantTri(A: Matrix) return Scalar;                 -- determinant of a tridiagonal matrix
  procedure CharPolyTri(A: in Matrix; P: in out Polynom1);          -- characteristic polynomial of a tridiagonal matrix
  procedure EigenPolyHess(A: in Matrix; E: in out EigenPolys);      -- E(0..N,*) are polynomials of degree <= N, for A(1..N,1..N)
  procedure GetCharPoly(E: in EigenPolys; P: in out Polynom1);      -- last row of E
  procedure EigenSymm(A: in Matrix; W: in out Vector);                    -- needs approx eigenvalues W
  procedure EigenSymm(A: in Matrix; W: in out Vector; V: in out Matrix);  -- needs approx eigenvalues W
  procedure EigenCircle(A: in Matrix; Wr: in out Vector);                 -- needs approx eigenvalues Wr (real part)
  procedure Re2Im(Wr: in Vector; Wi: in out Vector);                      -- compute imaginary parts
  procedure EigenCircle(A: in Matrix; Wr,Wi: in out Vector; Vr,Vi: in out Matrix); -- needs approx eigenvalues Wr (real part)

  --- these use Num_PolyRoots to find approx solutions (eps is for bisection)
  procedure Real_Root(P: in Polynom1; WMin,WMax: in Flt; W: in out Scalar; Eps: in Radius);  -- slow, avoid
  procedure Real_Roots(P: in Polynom1; WMin,WMax: in Flt; W: in out Vector; Eps: in Radius);
  procedure Circle_Roots(P: in Polynom1; W: in out Vector; Eps: in Radius);
  procedure EigenSymm(A: in Matrix; WMin,WMax: in Flt; W: in out Vector; Eps: in Radius);
  procedure EigenSymm(A: in Matrix; W: in out Vector; Eps: in Radius);                    -- determines WMin,WMax via MaxNorm2
  procedure EigenSymm(A: in Matrix; WMin,WMax: in Flt; W: in out Vector; V: in out Matrix; Eps: in Radius);
  procedure EigenSymm(A: in Matrix; W: in out Vector; V: in out Matrix; Eps: in Radius);  -- determines WMin,WMax via MaxNorm2
  procedure EigenCircle(A: in Matrix; Wr: in out Vector; Eps: in Radius);
  procedure EigenCircle(A: in Matrix; Wr,Wi: in out Vector; Vr,Vi: in out Matrix; Eps: in Radius);

  procedure MaxEigen(A,Tmp: in out Matrix; S: in out Scalar; Pow4: in Positive := 4);
  function MaxEigen(A: Matrix; Pow4: Positive := 4) return Scalar;         -- upper bound on largest |eigenvalue|
  function MinEigen(A: Matrix; Pow4: Positive := 4) return Scalar;         -- MaxEigen for 1/A
  procedure InvNewton(A: in Matrix; Ai: in out Matrix; Iter: in Positive); -- needs approx Ai
  procedure InvNewton(A: in out Matrix; Iter: in Positive);

  function GetDiag(A: in Matrix) return Diag;
  procedure Add(D: in Diag; A: in out Matrix);                     -- A := A+D
  procedure Sub(D: in Diag; A: in out Matrix);                     -- A := A-D
  procedure Mult(D: in Diag; W: in out Vector);                    -- W := D*W
  procedure Mult(D1: in Diag; D2: in out Diag);                    -- D2 := D1*D2
  procedure Mult(D: in Diag; A: in out Matrix);                    -- A := D*A
  procedure Mult(A: in out Matrix; D: in Diag);                    -- A := A*D
  function "*"(D: Diag; A: Matrix) return Matrix;                  -- return D*A
  function "*"(A: Matrix; D: Diag) return Matrix;                  -- return A*D
  procedure Invert(D: in out Diag);                                -- invert D
  function Inverse(D: in Diag) return Diag;                        -- return D^{-1}

  --- testing and info
  procedure RandomVec(V: in out Vector; Fac: in Flt := One);             -- random vector
  procedure RandomMat(A: in out Matrix; Fac: in Flt := One);             -- random matrix
  procedure RandomOrth(A: in out Matrix);                                -- random orthogonal matrix
  procedure RandomSymm(W: in out Vector; A: in out Matrix);              -- random symmetric A with eigenvalues W
  procedure RandomCircle(Wr: in out Vector; A: in out Matrix);           -- random cicle A with eigenvalues Wr+-i*Wi
  procedure CheckSymm(A: in Matrix);    -- show norm1 of A-A^*
  procedure CheckInv(A,Ai: in Matrix);  -- show norm1 of A*Ai-Id
  procedure CheckOrth(A: in Matrix);    -- show norm1 of A*A^*-Id

  procedure EigenQuality(A: in Matrix; W: in Scalar; V: in Vector);           -- norm1 of A*V-W*V
  procedure EigenQuality(A: in Matrix; Wr,Wi: in Scalar; Vr,Vi: in Vector);
  procedure EigenQuality(Ar,Ai: in Matrix; Wr,Wi: in Scalar; Vr,Vi: in Vector);
  procedure EigenQuality(A: in Matrix; W: in Vector; V: in Matrix);
  procedure EigenQuality(A: in Matrix; Wr,Wi: in Vector; Vr,Vi: in Matrix);
  procedure EigenQuality(Ar,Ai: in Matrix; Wr,Wi: in Vector; Vr,Vi: in Matrix);
  procedure Col_Normalize1(Vr,Vi: in out Matrix); -- approx

private

  type Pivot is array(Integer range <>) of Integer;

  procedure CheckDim(N: in Positive; V: in Vector);
  procedure CheckDim(N: in Positive; A: in Matrix);
  procedure CheckPolyDim(D: in Natural; P: in Polynom1);

  procedure Dgefa(A: in out Matrix; Pvt: out Pivot);
  procedure Dgesl(Pvt: in Pivot; A: in Matrix; B: in out Vector);
  procedure Dgedi(Pvt: in Pivot; A: in out Matrix; B: in out Vector);
  procedure HHLeftMult(U,V: in Vector; A: in out Matrix; Tmp: in out Vector);  -- A := Householder(U)*A, use V=U^*
  procedure HHRightMult(U,V: in Vector; A: in out Matrix; Tmp: in out Vector); -- A := A*Householder(U), use V=U^*
  procedure Re2Im(C: in Scalar; S,Tmp: in out Scalar);
  procedure Symplectic_Roots(P: in Polynom1; W: in out Vector; Eps: in Radius);
  procedure InvNewtonStep(A: in Matrix; Ai,B1,B2: in out Matrix; Err: out Radius);

  One_Scalar:   constant Scalar      := Scal(One);
  Half_Scalar:  constant Scalar      := Scal(Half);
  SInfo:        constant Scalar_Info := Info(One_Scalar);
  STrunc:       constant Boolean     := SInfo.IsNumeric;
  SComplex:     constant Boolean     := SInfo.IsComplex;

  pragma Inline (IsZero,SetZero,Scal,Neg,Add,Sub,Mult,AddProd,Adjoint);
  pragma Inline_Always (Swap,Neg,Add,Sub,Mult,AddProd,Scal,CheckDim);

  use SV_Ops;

end ScalVectors;
